#!/usr/bin/env python
# filename: alignment.py
#
#
# Copyright (c) 2023 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


import os
import platform
import subprocess as sp
import sys
import tempfile
import uuid
from abc import ABC, abstractmethod
from copy import copy, deepcopy
from io import StringIO
from itertools import groupby
from typing import Callable, Iterable, Optional, Union

import parasail
from Bio import AlignIO
from Bio.Align import AlignInfo, MultipleSeqAlignment
from Bio.SeqRecord import SeqRecord

from ..core.sequence import Sequence
from ..io import to_fasta
from .decorators import lazy_property

STR_TYPES = [str]

__all__ = [
    "mafft",
    "muscle",
    "muscle_v3",
    "local_alignment",
    "global_alignment",
    "semiglobal_alignment",
    "dot_alignment",
    "LocalAlignment",
    "GlobalAlignment",
    "SemiGlobalAlignment",
]


# -------------------------------------
#
#     MULTIPLE SEQUENCE ALIGNMENT
#
# -------------------------------------


def mafft(
    sequences: Union[str, Iterable],
    alignment_file: Optional[str] = None,
    fmt: str = "fasta",
    threads: int = -1,
    as_file: bool = False,
    as_string: bool = False,
    reorder: bool = True,
    mafft_bin: Optional[str] = None,
    id_key: Optional[str] = None,
    seq_key: Optional[str] = None,
    debug: bool = False,
    fasta: Optional[str] = None,
) -> Union[MultipleSeqAlignment, str]:
    """
    Performs multiple sequence alignment with `MAFFT`_.

    Parameters
    ----------
    sequences : (str, iterable)
        Can be one of several things:
            1. path to a FASTA-formatted file
            2. a FASTA-formatted string
            3. a list of BioPython ``SeqRecord`` objects
            4. a list of abutils ``Sequence`` objects
            5. a list of lists/tuples, of the format ``[sequence_id, sequence]``

    alignment_file : str, optional
        Path for the output alignment file. Required if ``as_file`` is ``True``.

    fmt : str, default='fasta'
        Format of the output alignment. Choices are 'fasta', 'phylip', and 'clustal'.
        Default is 'fasta'.

    threads : int, default=-1
        Number of threads for MAFFT to use. Default is ``-1``, which uses all
        available CPUs.

    as_file: bool, default=False
        If ``True``, returns the path to the alignment file. If ``False``,
        returns either a BioPython ``MultipleSeqAlignment`` object or the alignment
        output as a ``str``, depending on `as_string`. If `alignment_file` is not
        provided, a temporary file will be created with ``tempfile.NamedTemporaryFile``.
        Note that this temporary file is created in ``"/tmp"`` and may be removed
        by the operating system without notice.

    as_string: bool, default=False
        If ``True``, returns a the alignment output as a string. If ``False``,
        returns a BioPython ``MultipleSeqAlignment`` object (obtained by calling
        ``Bio.AlignIO.read()`` on the alignment file).

    mafft_bin : str, optional
        Path to a MAFFT executable. Default is ``None``, which results in
        using the default system MAFFT binary (just calling ``'mafft'``).

    id_key : str, default=None
        Key to retrieve the sequence ID. If not provided or missing, ``Sequence.id`` is used.

    sequence_key : str, default=None
        Key to retrieve the sequence. If not provided or missing, ``Sequence.sequence`` is used.

    debug : bool, default=False
        If ``True``, prints MAFFT's standard output and standard error.
        Default is ``False``.

    fasta : str, optional
        Path to a FASTA-formatted input file. Depricated (use `sequences`
        for all types if input), but retained for backwards compatibility.


    Returns
    -------
    alignment : str or ``MultipleSeqAlignment``
        If ``as_file`` is ``True``, returns a path to the output alignment file,
        Otherwise, returns a BioPython ``MultipleSeqAlignment`` object.


    .. _MAFFT:
        https://mafft.cbrc.jp/alignment/software/
    """
    # process input
    if fasta is not None:
        sequences = fasta
    ffile = to_fasta(sequences, id_key=id_key, sequence_key=seq_key)
    # configure output path
    if alignment_file is None:
        # as_file = False
        alignment_file = tempfile.NamedTemporaryFile(delete=False).name
    else:
        alignment_file = os.path.abspath(alignment_file)
    # do the alignment
    aln_format = ""
    if fmt.lower() == "clustal":
        aln_format = "--clustalout "
    if fmt.lower() == "phylip":
        aln_format = "--phylipout "
    if reorder:
        aln_format += "--reorder "
    if mafft_bin is None:
        mafft_bin = "mafft"
    mafft_cline = "{} --thread {} {}{} > {}".format(
        mafft_bin, threads, aln_format, ffile, alignment_file
    )
    mafft = sp.Popen(
        str(mafft_cline),
        stdout=sp.PIPE,
        stderr=sp.PIPE,
        universal_newlines=True,
        shell=True,
    )
    stdout, stderr = mafft.communicate()
    # return output
    if debug:
        print(mafft_cline)
        print(stdout)
        print(stderr)
    if os.stat(alignment_file).st_size == 0:
        err = "WARNING: output alignment file is empty. "
        err += "Verify that MAFFT is installed and the input data is valid,"
        print(err)
        return None
    if as_file:
        return alignment_file
    with open(alignment_file, "r") as f:
        if as_string:
            aln = f.read()
        else:
            aln = AlignIO.read(f, fmt)
    os.unlink(alignment_file)
    return aln


def muscle(
    sequences: Union[str, Iterable],
    alignment_file: Optional[str] = None,
    as_file: bool = False,
    as_string: bool = False,
    muscle_bin: Optional[str] = None,
    threads: Optional[int] = None,
    id_key: Optional[str] = None,
    seq_key: Optional[str] = None,
    debug: bool = False,
    fasta: Optional[str] = None,
) -> Union[MultipleSeqAlignment, str]:
    """
    Performs multiple sequence alignment with `MUSCLE`_ (version 5).

    Parameters
    ----------
    sequences : (str, iterable)
        Can be one of several things:
            1. path to a FASTA-formatted file
            2. a FASTA-formatted string
            3. a list of BioPython ``SeqRecord`` objects
            4. a list of abutils ``Sequence`` objects
            5. a list of lists/tuples, of the format ``[sequence_id, sequence]``

    alignment_file : str, optional
        Path for the output alignment file. Required if ``as_file`` is ``True``.

    as_file: bool, default=False
        If ``True``, returns the path to the alignment file. If ``False``,
        returns either a BioPython ``MultipleSeqAlignment`` object or the alignment
        output as a ``str``, depending on `as_string`. If `alignment_file` is not
        provided, a temporary file will be created with ``tempfile.NamedTemporaryFile``.
        Note that this temporary file is created in ``"/tmp"`` and may be removed
        by the operating system without notice.

    as_string: bool, default=False
        If ``True``, returns a the alignment output as a string. If ``False``,
        returns a BioPython ``MultipleSeqAlignment`` object (obtained by calling
        ``Bio.AlignIO.read()`` on the alignment file).

    muscle_bin : str, optional
        Path to a MUSCLE executable. If not provided, the MUSCLE binary bundled
        with ``abutils`` will be used.

    threads : int, default=None
        Number of threads for MUSCLE to use. If not provided, MUSCLE uses all
        available CPUs.

    id_key : str, default=None
        Key to retrieve the sequence ID. If not provided or missing, ``Sequence.id`` is used.

    sequence_key : str, default=None
        Key to retrieve the sequence. If not provided or missing, ``Sequence.sequence`` is used.

    debug : bool, default=False
        If ``True``, prints MAFFT's standard output and standard error.
        Default is ``False``.

    fasta : str, optional
        Path to a FASTA-formatted input file. Depricated (use `sequences`
        for all types if input), but retained for backwards compatibility.


    Returns
    -------
    alignment : str or ``MultipleSeqAlignment``
        If ``as_file`` is ``True``, returns a path to the output alignment file,
        Otherwise, returns a BioPython ``MultipleSeqAlignment`` object.


    .. _MUSCLE:
        https://www.drive5.com/muscle/
    """
    # process input
    if fasta is not None:
        sequences = fasta
    ffile = to_fasta(sequences, id_key=id_key, sequence_key=seq_key)
    # configure output path
    if alignment_file is None:
        # as_file = False
        alignment_file = tempfile.NamedTemporaryFile(delete=False).name
    else:
        alignment_file = os.path.abspath(alignment_file)
    # muscle binary
    if muscle_bin is None:
        mod_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        muscle_bin = os.path.join(
            mod_dir, "bin/muscle_{}".format(platform.system().lower())
        )
    # do the alignment
    muscle_cline = f"{muscle_bin} -align {ffile} -output {alignment_file}"
    if threads is not None:
        muscle_cline += f" -threads {threads}"
    muscle = sp.Popen(str(muscle_cline), stdout=sp.PIPE, stderr=sp.PIPE, shell=True)
    stdout, stderr = muscle.communicate()
    if debug:
        print("muscle binary path:", muscle_bin)
        print("muscle command:", muscle_cline)
        print(stdout)
        print(stderr)
    # output
    if os.stat(alignment_file).st_size == 0:
        err = "WARNING: output alignment file is empty. "
        err += "Please verify that the input data is valid"
        print(err)
        return None
    if as_file:
        return alignment_file
    with open(alignment_file, "r") as f:
        if as_string:
            aln = f.read()
        else:
            aln = AlignIO.read(f, "fasta")
    os.unlink(alignment_file)
    return aln


def muscle_v3(
    sequences: Union[str, Iterable, None] = None,
    alignment_file: Optional[str] = None,
    fmt: str = "fasta",
    as_file: bool = False,
    as_string: bool = False,
    muscle_bin: Optional[str] = None,
    maxiters: Optional[int] = None,
    diags: Optional[int] = None,
    gap_open: Optional[int] = None,
    gap_extend: Optional[int] = None,
    id_key: Optional[str] = None,
    seq_key: Optional[str] = None,
    debug: bool = False,
    fasta: Optional[str] = None,
) -> Union[MultipleSeqAlignment, str]:
    """
    Performs multiple sequence alignment with `MUSCLE`_ (version 3).

    Parameters
    ----------
    sequences : (str, iterable)
        Can be one of several things:
            1. path to a FASTA-formatted file
            2. a FASTA-formatted string
            3. a list of BioPython ``SeqRecord`` objects
            4. a list of abutils ``Sequence`` objects
            5. a list of lists/tuples, of the format ``[sequence_id, sequence]``

    alignment_file : str, optional
        Path for the output alignment file. Required if ``as_file`` is ``True``.

    fmt : str, default='fasta'
        Format of the output alignment. Choices are 'fasta', 'phylip', and 'clustal'.
        Default is 'fasta'.

    threads : int, default=-1
        Number of threads for MUSCLE to use. Default is ``-1``, which uses all
        available CPUs.

    as_file: bool, default=False
        If ``True``, returns the path to the alignment file. If ``False``,
        returns either a BioPython ``MultipleSeqAlignment`` object or the alignment
        output as a ``str``, depending on `as_string`. Requires that
        `alignment_file` is also provided.

    as_string: bool, default=False
        If ``True``, returns a the alignment output as a string. If ``False``,
        returns a BioPython ``MultipleSeqAlignment`` object (obtained by calling
        ``Bio.AlignIO.read()`` on the alignment file).

    maxiters : int, default=-1
        Passed directly to MUSCLE using the ``-maxiters`` flag.

    diags : int, default=None
        Passed directly to MUSCLE using the ``-diags`` flag.

    gap_open : float, default=None
        Passed directly to MUSCLE using the ``-gapopen`` flag. Ignored
        if ``gap_extend`` is not also provided.

    gap_extend : float, default=None
        Passed directly to MUSCLE using the ``-gapextend`` flag. Ignored
        if ``gap_open`` is not also provided.

    muscle_bin : str, optional
        Path to a MUSCLE executable. If not provided, the MUSCLE binary bundled
        with ``abutils`` will be used.

    debug : bool, default=False
        If ``True``, prints MAFFT's standard output and standard error.
        Default is ``False``.

    fasta : str, optional
        Path to a FASTA-formatted input file. Depricated (use `sequences`
        for all types if input), but retained for backwards compatibility.


    Returns
    -------
    alignment : str or ``MultipleSeqAlignment``
        If ``as_file`` is ``True``, returns a path to the output alignment file,
        Otherwise, returns a BioPython ``MultipleSeqAlignment`` object.


    .. _MUSCLE:
        https://drive5.com/muscle/downloads_v3.htm

    """
    # process input
    if fasta is not None:
        sequences = fasta
    fasta_string = to_fasta(
        sequences,
        as_string=True,
        id_key=id_key,
        sequence_key=seq_key,
    )
    # configure output path
    if alignment_file is None:
        as_file = False
        alignment_file = tempfile.NamedTemporaryFile(delete=False).name
    else:
        alignment_file = os.path.abspath(alignment_file)
    # muscle binary
    if muscle_bin is None:
        mod_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        muscle_bin = os.path.join(
            mod_dir, "bin/muscle3_{}".format(platform.system().lower())
        )
    # do the alignment
    aln_format = ""
    if fmt == "clustal":
        aln_format = " -clwstrict"
    muscle_cline = "{}{} ".format(muscle_bin, aln_format)
    if maxiters is not None:
        muscle_cline += " -maxiters {}".format(maxiters)
    if diags:
        muscle_cline += " -diags"
    if all([gap_open is not None, gap_extend is not None]):
        muscle_cline += " -gapopen {} -gapextend {}".format(gap_open, gap_extend)
    if debug:
        print("muscle binary path:", muscle_bin)
        print("muscle command:", muscle_cline)
    muscle = sp.Popen(
        str(muscle_cline),
        stdin=sp.PIPE,
        stdout=sp.PIPE,
        stderr=sp.PIPE,
        universal_newlines=True,
        shell=True,
    )
    alignment, stderr = muscle.communicate(input=fasta_string)
    if debug:
        print(stderr)
    # output
    if as_string:
        return alignment
    elif as_file:
        with open(alignment_file, "w") as f:
            f.write(alignment)
    else:
        return AlignIO.read(StringIO(alignment), fmt)


def consensus(aln, name=None, threshold=0.51, ambiguous="N"):
    summary_align = AlignInfo.SummaryInfo(aln)
    consensus = summary_align.gap_consensus(threshold=threshold, ambiguous=ambiguous)
    if name is None:
        name = uuid.uuid4()
    consensus_string = str(consensus).replace("-", "")
    return (name, consensus_string.upper())


# ----------------------------
#
#     PAIRWISE ALIGNMENT
#
# ----------------------------


class PairwiseAlignment(ABC):
    """
    Base class for local, global, and semi-global pairwise alignments.

    .. note::

        All comparisons between ``PairwiseAlignment``s are done on the
        ``score`` attribute. This was done so that sorting alignments
        like so::

            list_of_alignments.sort(reverse=True)

        produces a sorted list of alignments from the highest alignment
        score to the lowest.

    Attributes
    ----------

        query (Sequence): The input query sequence, as an AbTools
            ``Sequence`` object.

        target (Sequence): The input target sequence, as an AbTools
            ``Sequence`` object.

        target_id (str): ID of the target sequence.

        raw_query: The raw query, before conversion to a ``Sequence``.

        raw_target: The raw target, before conversion to a ``Sequence``.

    """

    def __init__(
        self,
        query: Sequence,
        target: Sequence,
        match: int = 3,
        mismatch: int = -2,
        gap_open: int = 5,
        gap_extend: int = 2,
        matrix: Optional[parasail.Matrix] = None,
        gap_open_penalty: Optional[int] = None,
        gap_extend_penalty: Optional[int] = None,
    ):
        self.query = query
        self.target = target
        self.match = int(match)
        self.mismatch = int(mismatch)
        self.gap_open = self.get_gap_open(gap_open, gap_open_penalty)
        self.gap_extend = self.get_gap_extend(gap_extend, gap_extend_penalty)
        self.matrix = self.get_matrix(matrix)
        self._alignment = None
        self._traceback = None
        self._cigar = None

        # for backwards compatibility:
        self.raw_query = query
        self.raw_target = target

    def __repr__(self):
        if len(self.aligned_query) > 20:
            qstring = "{}...{}".format(
                self.aligned_query[:10], self.aligned_query[-10:]
            )
            mstring = "{}...{}".format(
                self.alignment_midline[:10], self.alignment_midline[-10:]
            )
            tstring = "{}...{}".format(
                self.aligned_target[:10], self.aligned_target[-10:]
            )
        else:
            qstring = self.aligned_query
            mstring = self.alignment_midline
            tstring = self.aligned_target
        lines = []
        lines.append("Pairwise Alignment")
        lines.append("------------------\n")
        lines.append(f"query:  {qstring}")
        lines.append(f"        {mstring}")
        lines.append(f"target: {tstring}\n")
        lines.append(f"score: {self.score}")
        lines.append(f"type: {self.alignment_type}")
        lines.append(f"length: {len(self.aligned_query)}")
        print("\n".join(lines))
        return ""

    def __str__(self):
        lines = []
        lines.append(self.aligned_query)
        lines.append(self.alignment_midline)
        lines.append(self.aligned_target)
        return "\n".join(lines)

    def __len__(self):
        try:
            return len(self.aligned_query)
        except AttributeError:
            return 0

    def __eq__(self, other):
        if not hasattr(other, "score"):
            if type(other) in [int, float]:
                return self.score == other
            return False
        return self.score == other.score

    def __lt__(self, other):
        if not hasattr(other, "score"):
            if type(other) in [int, float]:
                return self.score < other
            return False
        return self.score < other.score

    def __le__(self, other):
        if not hasattr(other, "score"):
            if type(other) in [int, float]:
                return self.score <= other
            return False
        return self.score <= other.score

    def __gt__(self, other):
        if not hasattr(other, "score"):
            if type(other) in [int, float]:
                return self.score > other
            return False
        return self.score > other.score

    def __ge__(self, other):
        if not hasattr(other, "score"):
            if type(other) in [int, float]:
                return self.score >= other
            return False
        return self.score >= other.score

    @lazy_property
    def profile(self):
        return parasail.profile_create_16(
            self.query.sequence, self.match, self.mismatch, self.matrix
        )

    @lazy_property
    def alignment(self):
        return self.align()

    @property
    def cigar(self):
        cigar_string = self.alignment.cigar.decode.decode("utf-8")
        return CIGAR(cigar_string=cigar_string)

    @property
    def traceback(self):
        return self.alignment.traceback

    @lazy_property
    def score(self):
        return self.alignment.score

    @lazy_property
    def aligned_query(self):
        return self.traceback.query

    @lazy_property
    def aligned_target(self):
        return self.traceback.ref

    @lazy_property
    def alignment_midline(self):
        return self.traceback.comp.replace(".", " ")

    @lazy_property
    def query_begin(self):
        # return self.alignment.cigar.beg_query
        query_begin = self.alignment.cigar.beg_query
        if self.cigar[0].element == "I":
            query_begin += self.cigar[0].length
        return query_begin

    @lazy_property
    def query_end(self):
        return self.alignment.end_query

    @lazy_property
    def target_begin(self):
        # return self.alignment.cigar.beg_ref
        target_begin = self.alignment.cigar.beg_ref
        if self.cigar[0].element == "D":
            target_begin += self.cigar[0].length
        return target_begin

    @lazy_property
    def target_end(self):
        return self.alignment.end_ref

    @property
    def target_id(self):
        return self.target.id

    def align(self):
        """
        docstring for align
        """
        aln = self.alignment_function(
            self.query.sequence,
            self.target.sequence,
            self.gap_open,
            self.gap_extend,
            self.matrix,
        )
        return aln

    def get_gap_open(self, gap_open, gap_open_penalty):
        if gap_open_penalty is not None:
            gap_open = gap_open_penalty
        return abs(int(gap_open))

    def get_gap_extend(self, gap_extend, gap_extend_penalty):
        if gap_extend_penalty is not None:
            gap_extend = gap_extend_penalty
        return abs(int(gap_extend))

    def get_matrix(self, matrix):
        if matrix is None:
            residues = "ACDEFGHIKLMNPQRSTVWY"
            return parasail.matrix_create(residues, self.match, self.mismatch)
        elif isinstance(matrix, str):
            if os.path.isfile(matrix):
                try:
                    return parasail.Matrix(matrix)
                except:
                    err = (
                        f"\nERROR: the provided matrix file:\n{matrix}\ndoes not exist "
                    )
                    err += "or is improperly formatted. Please provide a valid matrix file.\n"
                    print(err)
                    sys.exit(1)
            else:
                try:
                    return parasail.Matrix(matrix.lower())
                except ValueError:
                    err = f"\nERROR: the provided matrix name {matrix} was not found. "
                    err += "Please provide a valid parasail matrix name.\n"
                    print(err)
                    sys.exit(1)
        elif isinstance(matrix, parasail.Matrix):
            return matrix
        else:
            err = "\nERROR: If supplied, matrix must be either the name of a built-in "
            err += "parasail matrix, the path to a properly formatted matrix file, "
            err += f"or a parasail.Matrix object. You supplied:\n{matrix}\n"
            print(err)
            sys.exit(1)


class LocalAlignment(PairwiseAlignment):
    """
    docstring for LocalAlignment
    """

    def __init__(
        self,
        query: Sequence,
        target: Sequence,
        match: int = 3,
        mismatch: int = -2,
        gap_open: int = 5,
        gap_extend: int = 2,
        matrix: Union[str, parasail.Matrix, None] = None,
        gap_open_penalty: Optional[int] = None,
        gap_extend_penalty: Optional[int] = None,
        alignment_function: Callable = parasail.sw_trace_striped_16,
    ):
        super().__init__(
            query=query,
            target=target,
            matrix=matrix,
            match=match,
            mismatch=mismatch,
            gap_open=gap_open,
            gap_extend=gap_extend,
            gap_open_penalty=gap_open_penalty,
            gap_extend_penalty=gap_extend_penalty,
        )
        self.alignment_function = alignment_function
        self.alignment_type = "local"


class GlobalAlignment(PairwiseAlignment):
    """
    docstring for LocalAlignment
    """

    def __init__(
        self,
        query: Sequence,
        target: Sequence,
        match: int = 3,
        mismatch: int = -2,
        gap_open: int = 5,
        gap_extend: int = 2,
        matrix: Union[str, parasail.Matrix, None] = None,
        gap_open_penalty: Optional[int] = None,
        gap_extend_penalty: Optional[int] = None,
        alignment_function: Callable = parasail.nw_trace_striped_16,
    ):
        super().__init__(
            query=query,
            target=target,
            matrix=matrix,
            match=match,
            mismatch=mismatch,
            gap_open=gap_open,
            gap_extend=gap_extend,
            gap_open_penalty=gap_open_penalty,
            gap_extend_penalty=gap_extend_penalty,
        )
        self.alignment_function = alignment_function
        self.alignment_type = "global"


class SemiGlobalAlignment(PairwiseAlignment):
    """
    docstring for LocalAlignment
    """

    def __init__(
        self,
        query: Sequence,
        target: Sequence,
        match: int = 3,
        mismatch: int = -2,
        gap_open: int = 5,
        gap_extend: int = 2,
        matrix: Union[str, parasail.Matrix, None] = None,
        gap_open_penalty: Optional[int] = None,
        gap_extend_penalty: Optional[int] = None,
        alignment_function: Callable = parasail.sg_trace_striped_16,
    ):
        super().__init__(
            query=query,
            target=target,
            matrix=matrix,
            match=match,
            mismatch=mismatch,
            gap_open=gap_open,
            gap_extend=gap_extend,
            gap_open_penalty=gap_open_penalty,
            gap_extend_penalty=gap_extend_penalty,
        )
        self.alignment_function = alignment_function
        self.alignment_type = "semi-global"


class CIGAR:
    """
    Parsable representation of a CIGAR string. The CIGAR parsing code was taken
    almost vertabim from the `Cigar`_ package. Init argument is a CIGAR string.
    Iterating over a ``CIGAR`` object yields ``CIGARElement``s. The length of
    a ``CIGAR`` object is the length of the alignment encoded by the CIGAR string,
    not the length of the CIGAR string itself::

        cig = CIGAR("99=2X99=")
        len(cig) # --> 200
        len(cig.cigar_string) # --> 8

    .. _Cigar
        https://github.com/brentp/cigar
    """

    def __init__(self, cigar_string):
        self.cigar_string = cigar_string
        self._cigar_list = None

    def __str__(self):
        return self.cigar_string

    def __repr__(self):
        print(self.cigar_string)
        return ""

    def __len__(self):
        return sum([c.length for c in self.cigar_list])

    def __getitem__(self, slice):
        return self.cigar_list[slice]

    def __iter__(self):
        for c in self.cigar_list:
            yield c

    @property
    def cigar_list(self):
        if self._cigar_list is None:
            cigar_list = []
            cig_iter = groupby(self.cigar_string, lambda c: c.isdigit())
            for _, n in cig_iter:
                ce = CIGARElement(
                    length=int("".join(n)), element="".join(next(cig_iter)[1])
                )
                cigar_list.append(ce)
            self._cigar_list = cigar_list
        return self._cigar_list


class CIGARElement:
    """
    Representation of a single CIGAR element.
    """

    def __init__(self, length, element):
        self.length: int = int(length)
        self.element: str = element


def process_targets(target, targets):
    if target is not None:
        return [Sequence(target)]
    return [Sequence(t) for t in targets]


def local_alignment(
    query: Union[str, SeqRecord, Sequence, Iterable],
    target: Union[str, SeqRecord, Sequence, Iterable, None] = None,
    targets: Optional[Iterable] = None,
    match: int = 3,
    mismatch: int = -2,
    gap_open: int = -5,
    gap_extend: int = -2,
    matrix: Union[str, parasail.Matrix, None] = None,
    alignment_function: Callable = parasail.sw_trace_striped_16,
    aa: bool = False,
    gap_open_penalty: Optional[int] = None,
    gap_extend_penalty: Optional[int] = None,
) -> Union[LocalAlignment, Iterable]:
    """
    Striped Smith-Waterman local pairwise sequence alignment.

    Parameters
    ----------
    query : str, SeqRecord, Sequence, or list
        Query sequence. Can be any of the following:
            1. a nucleotide or amino acid sequence, as a string
            2. a Biopython ``SeqRecord`` object
            3. an abutils ``Sequence`` object
            4. a ``list`` or ``tuple`` of the format ``[seq_id, sequence]``

    target : str, SeqRecord, Sequence, or list, default=None
        Target sequence. Can be anything accepted by `query`. One of
        `target` or `targets` must be provided.

    target : str, SeqRecord, Sequence, or list, default=None
        Iterable of multiple target sequences. A ``list`` or ``tuple`` of
        anything accepted by `query`. One of `target` or `targets` must be provided.

    match : int, default=3
        Match score. Should be a positive integer.

    mismatch : int, default=-2
        Mismatch score. Typically should be less than or equal to ``0``.

    gap_open : int, default=-5
        Gap open score. Must be less than or equal to ``0``.

    gap_extend : int, default=-2
        Gap extension score. Must be less than or equal to ``0``.

    matrix : str, dict or parasail.Matrix, default=None
        Scoring matrix. Can be either:
            - the name of a ``parasail`` `built-in substitution matrix`_.
            - path to a matrix file (in a format accepted by ``parasail``).
            - a ``parasail.Matrix`` object
        If not provided, a matrix will be created using `match` and `mismatch`.

    alignment_function : Callable, default=parasail.sw_trace_striped_16
        ``parasail`` `local alignment function`_ to be used for alignment.

    aa : bool, default=False
        Not used. Retained for backwards compatibility.

    gap_open_penalty : int, default=None
        Depricated, but retained for backwards compatibility. Use `gap_open`
        insteaed.

    gap_extend_penalty : int, default=None
        Depricated, but retained for backwards compatibility. Use `gap_exted`
        insteaed.


    Returns
    -------
    alignment : LocalAlignment or iterable
        If a single target sequence is provided (via ``target``), a single ``LocalAlignment``
        object will be returned. If multiple target sequences are supplied (via ``targets``),
        a list of ``LocalAlignment`` objects will be returned.


    .. _local alignment function
        https://github.com/jeffdaily/parasail-python#standard-function-naming-convention

    .. _built-in substitution matrix
        https://github.com/jeffdaily/parasail-python#substitution-matrices
    """
    # input processing
    query = Sequence(query)
    targets = process_targets(target, targets)
    # alignment
    alignments = []
    for target in targets:
        aln = LocalAlignment(
            query=query,
            target=target,
            match=match,
            mismatch=mismatch,
            gap_open=gap_open,
            gap_extend=gap_extend,
            matrix=matrix,
            alignment_function=alignment_function,
            gap_open_penalty=gap_open_penalty,
            gap_extend_penalty=gap_extend_penalty,
        )
        alignments.append(aln)
    if len(alignments) == 1:
        return alignments[0]
    return alignments


def global_alignment(
    query: Union[str, SeqRecord, Sequence, Iterable],
    target: Union[str, SeqRecord, Sequence, Iterable, None] = None,
    targets: Optional[Iterable] = None,
    match: int = 3,
    mismatch: int = -2,
    gap_open: int = -5,
    gap_extend: int = -2,
    matrix: Union[str, parasail.Matrix, None] = None,
    alignment_function: Callable = parasail.nw_trace_striped_16,
    aa: bool = False,
    gap_open_penalty: Optional[int] = None,
    gap_extend_penalty: Optional[int] = None,
) -> Union[GlobalAlignment, Iterable]:
    """
    Needleman-Wunch global pairwise sequence alignment.

    Parameters
    ----------
    query : str, SeqRecord, Sequence, or list
        Query sequence. Can be any of the following:
            1. a nucleotide or amino acid sequence, as a string
            2. a Biopython ``SeqRecord`` object
            3. an abutils ``Sequence`` object
            4. a ``list`` or ``tuple`` of the format ``[seq_id, sequence]``

    target : str, SeqRecord, Sequence, or list, default=None
        Target sequence. Can be anything accepted by `query`. One of
        `target` or `targets` must be provided.

    target : str, SeqRecord, Sequence, or list, default=None
        Iterable of multiple target sequences. A ``list`` or ``tuple`` of
        anything accepted by `query`. One of `target` or `targets` must be provided.

    match : int, default=3
        Match score. Should be a positive integer.

    mismatch : int, default=-2
        Mismatch score. Typically should be less than or equal to ``0``.

    gap_open : int, default=-5
        Gap open score. Must be  less than or equal to ``0``.

    gap_extend : int, default=-2
        Gap extension score. Must be less than or equal to ``0``.

    matrix : str, dict or parasail.Matrix, default=None
        Scoring matrix. Can be either:
            - the name of a ``parasail`` `built-in substitution matrix`_.
            - path to a matrix file (in a format accepted by ``parasail``).
            - a ``parasail.Matrix`` object
        If not provided, a matrix will be created using `match` and `mismatch`.

    alignment_function : Callable, default=parasail.sw_trace_striped_16
        ``parasail`` `global alignment function`_ to be used for alignment.

    aa : bool, default=False
        Not used. Retained for backwards compatibility.

    gap_open_penalty : int, default=None
        Depricated, but retained for backwards compatibility. Use `gap_open`
        insteaed.

    gap_extend_penalty : int, default=None
        Depricated, but retained for backwards compatibility. Use `gap_exted`
        insteaed.


    Returns
    -------
    alignment : GlobalAlignment or iterable
        If a single target sequence is provided (via ``target``), a single ``GlobalAlignment``
        object will be returned. If multiple target sequences are supplied (via ``targets``),
        a list of ``GlobalAlignment`` objects will be returned.


    .. _global alignment function
        https://github.com/jeffdaily/parasail-python#standard-function-naming-convention

    .. _built-in substitution matrix
        https://github.com/jeffdaily/parasail-python#substitution-matrices
    """
    # input processing
    query = Sequence(query)
    targets = process_targets(target, targets)
    # alignment
    alignments = []
    for target in targets:
        aln = GlobalAlignment(
            query=query,
            target=target,
            match=match,
            mismatch=mismatch,
            gap_open=gap_open,
            gap_extend=gap_extend,
            matrix=matrix,
            alignment_function=alignment_function,
            gap_open_penalty=gap_open_penalty,
            gap_extend_penalty=gap_extend_penalty,
        )
        alignments.append(aln)
    if len(alignments) == 1:
        return alignments[0]
    return alignments


def semiglobal_alignment(
    query: Union[str, SeqRecord, Sequence, Iterable],
    target: Union[str, SeqRecord, Sequence, Iterable, None] = None,
    targets: Optional[Iterable] = None,
    match: int = 3,
    mismatch: int = -2,
    gap_open: int = -5,
    gap_extend: int = -2,
    matrix: Union[str, parasail.Matrix, None] = None,
    alignment_function: Callable = parasail.sg_trace_striped_16,
    aa: bool = False,
    gap_open_penalty: Optional[int] = None,
    gap_extend_penalty: Optional[int] = None,
) -> Union[SemiGlobalAlignment, Iterable]:
    """
    Semi-global pairwise sequence alignment.

    Parameters
    ----------
    query : str, SeqRecord, Sequence, or list
        Query sequence. Can be any of the following:
            1. a nucleotide or amino acid sequence, as a string
            2. a Biopython ``SeqRecord`` object
            3. an abutils ``Sequence`` object
            4. a ``list`` or ``tuple`` of the format ``[seq_id, sequence]``

    target : str, SeqRecord, Sequence, or list, default=None
        Target sequence. Can be anything accepted by `query`. One of
        `target` or `targets` must be provided.

    target : str, SeqRecord, Sequence, or list, default=None
        Iterable of multiple target sequences. A ``list`` or ``tuple`` of
        anything accepted by `query`. One of `target` or `targets` must be provided.

    match : int, default=3
        Match score. Should be a positive integer.

    mismatch : int, default=-2
        Mismatch score. Typically should be less than or equal to ``0``.

    gap_open : int, default=-5
        Gap open score. Must be  less than or equal to ``0``.

    gap_extend : int, default=-2
        Gap extension score. Must be less than or equal to ``0``.

    matrix : str, dict or parasail.Matrix, default=None
        Scoring matrix. Can be either:
            - the name of a ``parasail`` `built-in substitution matrix`_.
            - path to a matrix file (in a format accepted by ``parasail``).
            - a ``parasail.Matrix`` object
        If not provided, a matrix will be created using `match` and `mismatch`.

    alignment_function : Callable, default=parasail.sw_trace_striped_16
        ``parasail`` `semi-global alignment function`_ to be used for alignment.

    aa : bool, default=False
        Not used. Retained for backwards compatibility.

    gap_open_penalty : int, default=None
        Depricated, but retained for backwards compatibility. Use `gap_open`
        insteaed.

    gap_extend_penalty : int, default=None
        Depricated, but retained for backwards compatibility. Use `gap_exted`
        insteaed.


    Returns
    -------
    alignment : SemiGlobalAlignment or iterable
        If a single target sequence is provided (via ``target``), a single ``SemiGlobalAlignment``
        object will be returned. If multiple target sequences are supplied (via ``targets``),
        a list of ``SemiGlobalAlignment`` objects will be returned.


    .. _semi-global alignment function
        https://github.com/jeffdaily/parasail-python#standard-function-naming-convention

    .. _built-in substitution matrix
        https://github.com/jeffdaily/parasail-python#substitution-matrices
    """
    # input processing
    query = Sequence(query)
    targets = process_targets(target, targets)
    # alignment
    alignments = []
    for target in targets:
        aln = SemiGlobalAlignment(
            query=query,
            target=target,
            match=match,
            mismatch=mismatch,
            gap_open=gap_open,
            gap_extend=gap_extend,
            matrix=matrix,
            alignment_function=alignment_function,
            gap_open_penalty=gap_open_penalty,
            gap_extend_penalty=gap_extend_penalty,
        )
        alignments.append(aln)
    if len(alignments) == 1:
        return alignments[0]
    return alignments


# def local_alignment(
#     query: Union[str, SeqRecord, Sequence, Iterable],
#     target: Union[str, SeqRecord, Sequence, Iterable, None] = None,
#     targets: Optional[Iterable] = None,
#     match: int = 3,
#     mismatch: int = -2,
#     gap_open: int = -5,
#     gap_extend: int = -2,
#     matrix: Union[str, dict, None] = None,
#     aa: bool = False,
#     gap_open_penalty: Optional[int] = None,
#     gap_extend_penalty: Optional[int] = None,
# ):
#     """
#     Striped Smith-Waterman local pairwise alignment.

#     Parameters
#     ----------
#     query : str, SeqRecord, Sequence, or list
#         Query sequence. Can be any of the following:
#             1. a nucleotide or amino acid sequence, as a string
#             2. a Biopython ``SeqRecord`` object
#             3. an abutils ``Sequence`` object
#             4. a ``list`` or ``tuple`` of the format ``[seq_id, sequence]``

#     target : str, SeqRecord, Sequence, or list, default=None
#         Target sequence. Can be anything accepted by `query`. One of
#         `target` or `targets` must be provided.

#     target : str, SeqRecord, Sequence, or list, default=None
#         Iterable of multiple target sequences. A ``list`` or ``tuple`` of
#         anything accepted by `query`. One of `target` or `targets` must be provided.

#     match : int, default=3
#         Match score. Should be a positive integer.

#     mismatch : int, default=-2
#         Mismatch score. Typically should be less than or equal to ``0``.

#     gap_open : int, default=-5
#         Gap open score. Typically should be less than or equal to ``0``.

#     gap_extend : int, default=-2
#         Gap extension score. Typically should be less than or equal to ``0``.

#     matrix : str, dict, default=None
#         Scoring matrix for amino acid alignments. Can be either:
#             - The name of a built-in matrix. Current options are
#             ``"blosum62"`` and ``"pam250"``.
#             - A nested dictionary, giving an alignment score for each residue
#             pair. Should be formatted such that retrieving the alignment score
#             for ``"A"`` and ``"G"`` is accomplished by::
#                 matrix['A']['G']
#         If not provided, ``"blosum62"`` is used.

#     aa : bool, default=False
#         Must be set to ``True`` if sequences are amino acids.

#     gap_open_penalty : int, default=None
#         Depricated, but retained for backwards compatibility. Use `gap_open`
#         insteaed. Should be a positive integer.

#     gap_extend_penalty : int, default=None
#         Depricated, but retained for backwards compatibility. Use `gap_exted`
#         insteaed. Should be a positive integer.


#     Returns
#     -------
#     alignment : SSWAlignment or iterable
#         If a single target sequence is provided (via ``target``), a single ``SSWAlignment``
#         object will be returned. If multiple target sequences are supplied (via ``targets``),
#         a list of ``SSWAlignment`` objects will be returned.
#     """
#     if aa and matrix is None:
#         matrix = "blosum62"
#     if not any([target, targets]):
#         err = "\nERROR: You must supply either ``target`` or ``targets``.\n"
#         raise RuntimeError(err)
#     if target:
#         targets = [
#             target,
#         ]
#     # to maintain backward compatibility with earlier abutils API
#     if gap_open_penalty is not None:
#         gap_open = -gap_open_penalty
#     if gap_extend_penalty is not None:
#         gap_extend = -gap_extend_penalty
#     # do the alignment
#     alignments = []
#     for t in targets:
#         try:
#             alignment = SSWAlignment(
#                 query=query,
#                 target=t,
#                 match=match,
#                 mismatch=mismatch,
#                 matrix=matrix,
#                 gap_open=-gap_open,
#                 gap_extend=-gap_extend,
#                 aa=aa,
#             )
#             alignments.append(alignment)
#         except IndexError:
#             continue
#     if len(alignments) == 1:
#         return alignments[0]
#     return alignments


# def local_alignment_biopython(
#     query,
#     target=None,
#     targets=None,
#     match=3,
#     mismatch=-2,
#     matrix=None,
#     gap_open=-5,
#     gap_extend=-2,
#     aa=False,
# ):
#     if not target and not targets:
#         err = "ERROR: You must supply a target sequence (or sequences)."
#         raise RuntimeError(err)
#     if target:
#         targets = [
#             target,
#         ]
#     alignments = []
#     for t in targets:
#         try:
#             alignment = alignment = BiopythonAlignment(
#                 query=query,
#                 target=t,
#                 match=match,
#                 mismatch=mismatch,
#                 matrix=matrix,
#                 gap_open=gap_open,
#                 gap_extend=gap_extend,
#                 aa=aa,
#             )
#             alignments.append(alignment)
#         except IndexError:
#             continue
#     if len(alignments) == 1:
#         return alignments[0]
#     return alignments


# def global_alignment(
#     query,
#     target=None,
#     targets=None,
#     match=3,
#     mismatch=-2,
#     gap_open=-5,
#     gap_extend=-2,
#     score_match=None,
#     score_mismatch=None,
#     score_gap_open=None,
#     score_gap_extend=None,
#     matrix=None,
#     aa=False,
# ):
#     """
#     Needleman-Wunch global pairwise alignment.

#     With ``global_alignment``, you can score an alignment using different
#     paramaters than were used to compute the alignment. This allows you to
#     compute pure identity scores (match=1, mismatch=0) on pairs of sequences
#     for which those alignment parameters would be unsuitable. For example::

#         seq1 = 'ATGCAGC'
#         seq2 = 'ATCAAGC'

#     using identity scoring params (match=1, all penalties are 0) for both alignment
#     and scoring produces the following alignment::

#         ATGCA-GC
#         || || ||
#         AT-CAAGC

#     with an alignment score of 6 and an alignment length of 8 (identity = 75%). But
#     what if we want to calculate the identity of a gapless alignment? Using::

#         global_alignment(seq1, seq2,
#                          gap_open=20,
#                          score_match=1,
#                          score_mismatch=0,
#                          score_gap_open=10,
#                          score_gap_extend=1)

#     we get the following alignment::

#         ATGCAGC
#         ||  |||
#         ATCAAGC

#     which has an score of 5 and an alignment length of 7 (identity = 71%). Obviously,
#     this is an overly simple example (it would be much easier to force gapless alignment
#     by just iterating over each sequence and counting the matches), but there are several
#     real-life cases in which different alignment and scoring paramaters are desirable.

#     Args:

#         query: Query sequence. ``query`` can be one of four things:

#             1. a nucleotide or amino acid sequence, as a string

#             2. a Biopython ``SeqRecord`` object

#             3. an AbTools ``Sequence`` object

#             4. a list/tuple of the format ``[seq_id, sequence]``

#         target: A single target sequence. ``target`` can be anything that
#             ``query`` accepts.

#         targets (list): A list of target sequences, to be proccssed iteratively.
#             Each element in the ``targets`` list can be anything accepted by
#             ``query``.

#         match (int): Match score for alignment. Should be a positive integer. Default is 3.

#         mismatch (int): Mismatch score for alignment. Should be a negative integer. Default is -2.

#         gap_open (int): Penalty for opening gaps in alignment. Should be a negative integer.
#             Default is -5.

#         gap_extend (int): Penalty for extending gaps in alignment. Should be a negative
#             integer. Default is -2.

#         score_match (int): Match score for scoring the alignment. Should be a positive integer.
#             Default is to use the score from ``match`` or ``matrix``, whichever is provided.

#         score_mismatch (int): Mismatch score for scoring the alignment. Should be a negative
#             integer. Default is to use the score from ``mismatch`` or ``matrix``, whichever
#             is provided.

#         score_gap_open (int): Gap open penalty for scoring the alignment. Should be a negative
#             integer. Default is to use ``gap_open``.

#         score_gap_extend (int): Gap extend penalty for scoring the alignment. Should be a negative
#             integer. Default is to use ``gap_extend``.

#         matrix (str, dict): Alignment scoring matrix. Two options for passing the alignment matrix:

#             - The name of a built-in matrix. Current options are ``blosum62`` and ``pam250``.

#             - A nested dictionary, giving an alignment score for each residue pair. Should be
#               formatted such that retrieving the alignment score for A and G is accomplished by::

#                 matrix['A']['G']

#         aa (bool): Must be set to ``True`` if aligning amino acid sequences. Default
#             is ``False``.

#     Returns:

#         If a single target sequence is provided (via ``target``), a single ``NWAlignment``
#         object will be returned. If multiple target sequences are supplied (via ``targets``),
#         a list of ``NWAlignment`` objects will be returned.
#     """
#     if not target and not targets:
#         err = "ERROR: You must supply a target sequence (or sequences)."
#         raise RuntimeError(err)
#     if target:
#         targets = [
#             target,
#         ]
#     if type(targets) not in (list, tuple):
#         err = "ERROR: ::targets:: requires an iterable (list or tuple)."
#         err += "For a single sequence, use ::target::"
#         raise RuntimeError(err)
#     alignments = []
#     for t in targets:
#         alignment = NWAlignment(
#             query=query,
#             target=t,
#             match=match,
#             mismatch=mismatch,
#             gap_open=gap_open,
#             gap_extend=gap_extend,
#             score_match=score_match,
#             score_mismatch=score_mismatch,
#             score_gap_open=score_gap_open,
#             score_gap_extend=score_gap_extend,
#             matrix=matrix,
#             aa=aa,
#         )
#         alignments.append(alignment)
#     if target is not None:
#         return alignments[0]
#     return alignments


# class BaseAlignment(object):
#     """
#     Base class for local and global pairwise alignments.

#     .. note::

#         All comparisons between ``BaseAlignments``
#         are done on the ``score`` attribute (which must be implemented
#         by any classes that subclass ``BaseAlignment``). This was done
#         so that sorting alignments like so::

#             alignments = [list of alignments]
#             alignments.sort(reverse=True)

#         results in a sorted list of alignments from the highest alignment
#         score to the lowest.

#     Attributes:

#         query (Sequence): The input query sequence, as an AbTools
#             ``Sequence`` object.

#         target (Sequence): The input target sequence, as an AbTools
#             ``Sequence`` object.

#         target_id (str): ID of the target sequence.

#         raw_query: The raw query, before conversion to a ``Sequence``.

#         raw_target: The raw target, before conversion to a ``Sequence``.

#     """

#     def __init__(
#         self, query, target, matrix, match, mismatch, gap_open, gap_extend, aa
#     ):
#         super(BaseAlignment, self).__init__()
#         self.raw_query = query
#         self.raw_target = target
#         self.query = self._process_sequence(query, aa=aa)
#         self.target = self._process_sequence(target, aa=aa)
#         self._matrix = matrix
#         self._match = int(match)
#         self._mismatch = int(mismatch)
#         self._gap_open = int(gap_open)
#         self._gap_extend = int(gap_extend)
#         self._aa = bool(aa)

#     def __repr__(self):
#         if len(self.aligned_query) > 20:
#             qstring = "{}...{}".format(
#                 self.aligned_query[:10], self.aligned_query[-10:]
#             )
#             mstring = "{}...{}".format(
#                 self.alignment_midline[:10], self.alignment_midline[-10:]
#             )
#             tstring = "{}...{}".format(
#                 self.aligned_target[:10], self.aligned_target[-10:]
#             )
#         else:
#             qstring = self.aligned_query
#             mstring = self.alignment_midline
#             tstring = self.aligned_target
#         return_string = "\n\n"
#         return_string += "Pairwise Alignment\n"
#         return_string += "------------------\n\n"
#         return_string += "query:  {}\n".format(qstring)
#         return_string += "        {}\n".format(mstring)
#         return_string += "target: {}\n\n".format(tstring)
#         return_string += "score: {}\n".format(str(self.score))
#         return_string += "type: {}\n".format(self.alignment_type)
#         return_string += "length: {}".format(str(len(self.aligned_query)))
#         print(return_string)
#         return ""

#     def __str__(self):
#         return_string = ""
#         return_string += "{}\n".format(self.aligned_query)
#         return_string += "{}\n".format(self.alignment_midline)
#         return_string += "{}\n".format(self.aligned_target)
#         return return_string

#     def __len__(self):
#         return len(self.aligned_query)

#     def __eq__(self, other):
#         if not hasattr(other, "score"):
#             if type(other) in [int, float]:
#                 return self.score == other
#             return False
#         return self.score == other.score

#     def __lt__(self, other):
#         if not hasattr(other, "score"):
#             if type(other) in [int, float]:
#                 return self.score == other
#             return False
#         return self.score < other.score

#     def __le__(self, other):
#         if not hasattr(other, "score"):
#             if type(other) in [int, float]:
#                 return self.score == other
#             return False
#         return self.score <= other.score

#     def __gt__(self, other):
#         if not hasattr(other, "score"):
#             if type(other) in [int, float]:
#                 return self.score == other
#             return False
#         return self.score > other.score

#     def __ge__(self, other):
#         if not hasattr(other, "score"):
#             if type(other) in [int, float]:
#                 return self.score == other
#             return False
#         return self.score >= other.score

#     @property
#     def target_id(self):
#         return self._target_id

#     @target_id.setter
#     def target_id(self, target_id):
#         self._target_id = target_id

#     @property
#     def alignment_midline(self):
#         if self._alignment_midline is None:
#             self._alignment_midline = self._get_alignment_midline()
#         return self._alignment_midline

#     @staticmethod
#     def _process_sequence(sequence, aa):
#         if type(sequence) == Sequence:
#             return sequence
#         return Sequence(sequence)

#     def _get_alignment_midline(self):
#         midline = ""
#         for q, t in zip(self.aligned_query, self.aligned_target):
#             if q == t:
#                 midline += "|"
#             else:
#                 midline += " "
#         return midline


# class SSWAlignment(BaseAlignment):
#     """
#     Structure for performing and analyzing a Smith-Waterman local alignment.

#     .. note:

#         Exposed attributes and methods are the same as ``NWAlignment``, so
#         local and global alignmnts can be handled in the same way. In fact,
#         since comparisons are made based on score, local and global alignments
#         can be directly compared with constructions like::

#             local_aln == global_aln
#             local_aln > global_aln
#             alignments = sorted([global_aln, local_aln])

#     Attributes:

#         alignment_type (str): Is 'local' for all ``SSWAlignment`` objects.

#         aligned_query (str): The aligned query sequence (including gaps).

#         aligned_target (str): The aligned target sequence (including gaps).

#         alignment_midline (str): Midline for the aligned sequences, with ``|`` indicating
#           matches and a gap indicating mismatches::

#             print(aln.aligned_query)
#             print(aln.alignment_midline)
#             print(aln.aligned_target)

#             # ATGC
#             # || |
#             # ATCC

#         score (int): Alignment score.

#         query_begin (int): Position in the raw query sequence at which
#             the optimal alignment begins.

#         query_end (int): Position in the raw query sequence at which the
#             optimal alignment ends.

#         target_begin (int): Position in the raw target sequence at which
#             the optimal alignment begins.

#         target_end (int): Position in the raw target sequence at which the
#             optimal alignment ends.
#     """

#     def __init__(
#         self,
#         query,
#         target,
#         match=3,
#         mismatch=-2,
#         matrix=None,
#         gap_open=5,
#         gap_extend=2,
#         aa=False,
#     ):
#         super(SSWAlignment, self).__init__(
#             query, target, matrix, match, mismatch, gap_open, gap_extend, aa
#         )

#         self.alignment_type = "local"
#         self._alignment = self._align()
#         self._aligned_query = self._alignment.aligned_query_sequence
#         self._aligned_target = self._alignment.aligned_target_sequence
#         self._alignment_midline = None
#         self.score = self._alignment.optimal_alignment_score
#         self.cigar = self._alignment.cigar
#         self.query_begin = self._alignment.query_begin
#         self.query_end = self._alignment.query_end
#         self.target_begin = self._alignment.target_begin
#         self.target_end = self._alignment.target_end_optimal
#         self._alignment = None

#     @property
#     def alignment(self):
#         return self._alignment

#     @property
#     def aligned_query(self):
#         return self._aligned_query

#     @aligned_query.setter
#     def aligned_query(self, aligned_query):
#         self._aligned_query = aligned_query
#         self._alignment_midline = None

#     @property
#     def aligned_target(self):
#         return self._aligned_target

#     @aligned_target.setter
#     def aligned_target(self, aligned_target):
#         self._aligned_target = aligned_target
#         self._alignment_midline = None

#     def _align(self):
#         aligner = StripedSmithWaterman(
#             self.query.sequence,
#             match_score=self._match,
#             mismatch_score=self._mismatch,
#             gap_open_penalty=self._gap_open,
#             gap_extend_penalty=self._gap_extend,
#             substitution_matrix=self._matrix,
#             protein=self._aa,
#         )
#         return aligner(self.target.sequence)


# class BiopythonAlignment(BaseAlignment):
#     def __init__(
#         self,
#         query,
#         target,
#         match=3,
#         mismatch=-2,
#         matrix=None,
#         gap_open=5,
#         gap_extend=2,
#         aa=False,
#     ):
#         super(BiopythonAlignment, self).__init__(
#             query, target, matrix, match, mismatch, gap_open, gap_extend, aa
#         )

#         self.alignment_type = "local"
#         self._aln = self._align()
#         aln_query, aln_target, score, begin, end = self._aln
#         self._aligned_query = aln_query[begin:end]
#         self._aligned_target = aln_target[begin:end]
#         self._alignment_midline = None
#         self.score = score
#         self.query_begin = self._get_begin_pos(aln_query, begin)
#         self.query_end = self._get_end_pos(aln_query, end)
#         self.target_begin = self._get_begin_pos(aln_target, begin)
#         self.target_end = self._get_end_pos(aln_target, end)

#     @property
#     def alignment(self):
#         return self._aln

#     @property
#     def aligned_query(self):
#         return self._aligned_query

#     @aligned_query.setter
#     def aligned_query(self, aligned_query):
#         self._aligned_query = aligned_query
#         self._alignment_midline = None

#     @property
#     def aligned_target(self):
#         return self._aligned_target

#     @aligned_target.setter
#     def aligned_target(self, aligned_target):
#         self._aligned_target = aligned_target
#         self._alignment_midline = None

#     def _align(self):
#         aln = pairwise2.align.localms(
#             self.query.sequence,
#             self.target.sequence,
#             self._match,
#             self._mismatch,
#             self._gap_open,
#             self._gap_extend,
#         )
#         return aln[0]

#     def _get_begin_pos(self, seq, begin):
#         dashes = seq.count("-", 0, begin)
#         return begin - dashes

#     def _get_end_pos(self, seq, end):
#         return len(seq[:end].replace("-", ""))


# class NWAlignment(BaseAlignment):
#     """
#     Structure for performing and analyzing a Needleman-Wunch global alignment.

#     .. note:

#         Exposed attributes and methods are the same as ``SSWAlignment``, so
#         local and global alignmnts can be handled in the same way. In fact,
#         since comparisons are made based on score, local and global alignments
#         can be directly compared with constructions like::

#             local_aln == global_aln
#             local_aln > global_aln
#             alignments = sorted([global_aln, local_aln])

#     Attributes:

#         alignment_type (str): Is 'global' for all ``NWAlignment`` objects.

#         aligned_query (str): The aligned query sequence (including gaps).

#         aligned_target (str): The aligned target sequence (including gaps).

#         alignment_midline (str): Midline for the aligned sequences, with
#           ``|`` indicating matches and a gap indicating mismatches::

#             print(aln.aligned_query)
#             print(aln.alignment_midline)
#             print(aln.aligned_target)

#             # ATGC
#             # || |
#             # ATCC

#         score (int): Alignment score.

#         query_begin (int): Position in the raw query sequence at which
#             the optimal alignment begins.

#         query_end (int): Position in the raw query sequence at which the
#             optimal alignment ends.

#         target_begin (int): Position in the raw target sequence at which
#             the optimal alignment begins.

#         target_end (int): Position in the raw target sequence at which the
#             optimal alignment ends.
#     """

#     def __init__(
#         self,
#         query,
#         target,
#         match=3,
#         mismatch=-2,
#         gap_open=-5,
#         gap_extend=-2,
#         score_match=None,
#         score_mismatch=None,
#         score_gap_open=None,
#         score_gap_extend=None,
#         matrix=None,
#         aa=False,
#     ):
#         super(NWAlignment, self).__init__(
#             query, target, matrix, match, mismatch, gap_open, gap_extend, aa
#         )
#         self.alignment_type = "global"
#         self._score_match = int(score_match) if score_match is not None else None
#         self._score_mismatch = (
#             int(score_mismatch) if score_mismatch is not None else None
#         )
#         self._score_gap_open = (
#             int(score_gap_open) if score_gap_open is not None else None
#         )
#         self._score_gap_extend = (
#             int(score_gap_extend) if score_gap_extend is not None else None
#         )
#         self._matrix = matrix
#         self._alignment = self._align()
#         self._aligned_query = self._alignment[0]
#         self._aligned_target = self._alignment[1]
#         self._alignment_midline = None
#         self.score = self._score_alignment()

#     @property
#     def alignment(self):
#         return self._aln

#     @property
#     def aligned_query(self):
#         return self._aligned_query

#     @aligned_query.setter
#     def aligned_query(self, aligned_query):
#         self._aligned_query = aligned_query
#         self._alignment_midline = None

#     @property
#     def aligned_target(self):
#         return self._aligned_target

#     @aligned_target.setter
#     def aligned_target(self, aligned_target):
#         self._aligned_target = aligned_target
#         self._alignment_midline = None

#     def _get_matrix_file(self, match=None, mismatch=None, matrix=None):
#         matrix_dir = os.path.join(
#             os.path.dirname(os.path.abspath(__file__)), "matrices"
#         )
#         builtins = [os.path.basename(f) for f in list_files(matrix_dir)]
#         if matrix is None:
#             matrix_name = "match{}mismatch{}".format(abs(match), abs(mismatch))
#             if matrix_name not in builtins:
#                 self._build_matrix_from_params(
#                     match, mismatch, os.path.join(matrix_dir, matrix_name)
#                 )
#             return os.path.join(matrix_dir, matrix_name)
#         elif isinstance(matrix, dict):
#             dhash = hashlib.md5()
#             encoded = json.dumps(matrix, sort_keys=True).encode()
#             dhash.update(encoded)
#             matrix_name = dhash.hexdigest()
#             if matrix_name not in builtins:
#                 self._build_matrix_from_dict(
#                     matrix, os.path.join(matrix_dir, matrix_name)
#                 )
#             return os.path.join(matrix_dir, matrix_name)
#         else:
#             if self._matrix.lower() in builtin_names:
#                 return os.path.join(matrix_dir, self._matrix.lower())
#             else:
#                 err = "The supplied matrix name ({}) does not exist. ".format(matrix)
#                 err += "Built-in matrices are: {}".format("\n".join(builtins))
#                 raise RuntimeError(err)

#         # if matrix is not None:
#         #     matrix_name = matrix
#         # else:
#         #     matrix_name = 'match{}mismatch{}'.format(abs(match), abs(mismatch))
#         # if matrix_name.lower() in builtins:
#         #     return os.path.join(matrix_dir, matrix_name)
#         # builtin_names = [os.path.basename(f) for f in list_files(matrix_dir)]
#         # if self._matrix is not None:
#         #     if self._matrix.lower() in builtin_names:
#         #         return os.path.join(matrix_dir, self._matrix.lower())
#         #     else:
#         #         err = 'The supplied matrix name ({}) does not exist. '.format(matrix)
#         #         err += 'Built-in matrices are: {}'.format(', '.join(builtins))
#         #         raise RuntimeError(err)
#         # else:
#         #     self._build_matrix_from_params(match, mismatch, os.path.join(matrix_dir, matrix_name))
#         #     return os.path.join(matrix_dir, matrix_name)

#     def _align(self):
#         matrix = self._get_matrix_file(
#             match=self._match, mismatch=self._mismatch, matrix=self._matrix
#         )
#         aln = nw.global_align(
#             self.query.sequence,
#             self.target.sequence,
#             gap_open=self._gap_open,
#             gap_extend=self._gap_extend,
#             matrix=matrix,
#         )
#         return aln

#     def _score_alignment(self):
#         if all([self._score_match is not None, self._score_mismatch is not None]):
#             matrix = self._get_matrix_file(
#                 match=self._score_match, mismatch=self._score_mismatch
#             )
#         elif self._matrix is not None:
#             matrix = self._get_matrix_file(matrix=self._matrix)
#         else:
#             matrix = self._get_matrix_file(match=self._match, mismatch=self._mismatch)
#         gap_open = (
#             self._score_gap_open if self._score_gap_open is not None else self._gap_open
#         )
#         gap_extend = (
#             self._score_gap_extend
#             if self._score_gap_extend is not None
#             else self._gap_extend
#         )
#         aln = nw.score_alignment(
#             self.aligned_query,
#             self.aligned_target,
#             gap_open=gap_open,
#             gap_extend=gap_extend,
#             matrix=matrix,
#         )
#         return aln

#     @staticmethod
#     def _build_matrix_from_params(match, mismatch, matrix_file):
#         mstring = " {}".format(match) if len(str(match)) == 1 else str(match)
#         mmstring = " {}".format(mismatch) if len(str(mismatch)) == 1 else str(mismatch)
#         residues = [
#             "A",
#             "C",
#             "D",
#             "E",
#             "F",
#             "G",
#             "H",
#             "I",
#             "K",
#             "L",
#             "M",
#             "N",
#             "P",
#             "Q",
#             "R",
#             "S",
#             "T",
#             "V",
#             "W",
#             "Y",
#             "*",
#         ]
#         header = "   " + "  ".join(residues)
#         matlist = [
#             header,
#         ]
#         for r1 in residues:
#             resline = [
#                 r1,
#             ]
#             for r2 in residues:
#                 resline.append(mstring if r1 == r2 else mmstring)
#             matlist.append(" ".join(resline))
#         open(matrix_file, "w").write("\n".join(matlist))
#         return matrix_file

#     @staticmethod
#     def _build_matrix_from_dict(matrix, matrix_file):
#         residues = sorted(matrix.keys())
#         header = "   " + "  ".join(residues)
#         matlist = [
#             header,
#         ]
#         for r1 in residues:
#             resline = [
#                 r1,
#             ]
#             for r2 in residues:
#                 s = matrix[r1][r2]
#                 score = " {}".format(s) if len(str(s)) == 1 else str(s)
#                 resline.append(score)
#             matlist.append(" ".join(resline))
#         with open(matrix_file, "w") as f:
#             f.write("\n".join(matlist))
#         return matrix_file

#     @staticmethod
#     def _get_builtin_matrix(matrix_name):
#         matrix_dir = os.path.join(
#             os.path.dirname(os.path.abspath(__file__)), "matrices"
#         )
#         matrices = [os.path.basename(f) for f in list_files(matrix_dir)]
#         if matrix_name.lower() not in matrices:
#             err = "The maxtrix name you provided ({}) is not built-in.".format(
#                 matrix_name
#             )
#             err += "Built in matrices are: {}".format(", ".join(matrices))
#             raise RuntimeError()
#         return os.path.join(matrix_dir, matrix_name.lower())


# -------------------------------------
#
#       VISUAL ALIGNMENT
#
# -------------------------------------


def dot_alignment(
    sequences,
    seq_field=None,
    name_field=None,
    root=None,
    root_name=None,
    cluster_threshold=0.75,
    as_fasta=False,
    just_alignment=False,
):
    """
    Creates a dot alignment (dots indicate identity, mismatches are represented by the mismatched
    residue) for a list of sequences.

    Parameters
    -----------
    sequences : Iterable
        A list of Sequence objects to be aligned.

    seq_field : str
        Name of the sequence field key. Default is ``vdj_nt``.

    name_field : str
        Name of the name field key. Default is ``seq_id``.

    root : str, Sequence
        The sequence used to 'root' the alignment. This sequence will be at the top of the alignment
        and is the sequence against which dots (identity) will be evaluated. Can be provided either
        as a string corresponding to the name of one of the sequences in ``sequences`` or as a
        Sequence object. If not provided, ``sequences`` will be clustered at ``cluster_threshold``
        and the centroid of the largest cluster will be used.

    root_name : str
        Name of the root sequence. If not provided, the existing name of the root sequence
        (``name_field``) will be used. If ``root`` is not provided, the default ``root_name`` is
        ``'centroid'``.

    cluster_threshold : float
        Threshold with which to cluster sequences if ``root`` is not provided. Default is ``0.75``.

    as_fasta : bool
        If ``True``, returns the dot alignment as a FASTA-formatted string, rather than a string
        formatted for human readability

    just_alignment : bool
        If ``True``, returns just the dot-aligned sequences as a list.

    Returns
    --------
    If ``just_alignment`` is ``True``, a list of dot-aligned sequences (without sequence names) will
    be returned. If ``as_fasta`` is ``True``, a string containing the dot-aligned sequences in FASTA
    format will be returned. Otherwise, a formatted string containing the aligned sequences (with
    sequence names) will be returned.

    """
    import abstar

    from .cluster import cluster

    sequences = deepcopy(sequences)
    root = copy(root)

    # if custom seq_field is specified, copy to the .alignment_sequence attribute
    if seq_field is not None:
        if not all([seq_field in list(s.annotations.keys()) for s in sequences]):
            print(
                "\nERROR: {} is not present in all of the supplied sequences.\n".format(
                    seq_field
                )
            )
            sys.exit(1)
        for s in sequences:
            s.alignment_sequence = s[seq_field]
    else:
        for s in sequences:
            s.alignment_sequence = s.sequence

    # if custom name_field is specified, copy to the .id attribute
    if name_field is not None:
        if not all([name_field in list(s.annotations.keys()) for s in sequences]):
            print(
                "\nERROR: {} is not present in all of the supplied sequences.\n".format(
                    name_field
                )
            )
            sys.exit(1)
        for s in sequences:
            s.alignment_id = s[name_field]
    else:
        for s in sequences:
            s.alignment_id = s.id

    # parse the root sequence
    if all([root is None, root_name is None]):
        clusters = cluster(sequences, threshold=cluster_threshold, quiet=True)
        clusters = sorted(clusters, key=lambda x: x.size, reverse=True)
        centroid = clusters[0].centroid
        root = abstar.run(("centroid", centroid.sequence))
        root.alignment_id = "centroid"
        root.alignment_sequence = root[seq_field]
    elif type(root) in STR_TYPES:
        root = [s for s in sequences if s.alignment_id == root][0]
        if not root:
            print(
                "\nERROR: The name of the root sequence ({}) was not found in the list of input sequences.".format(
                    root
                )
            )
            print("\n")
            sys.exit(1)
        sequences = [s for s in sequences if s.alignment_id != root.alignment_id]
    elif type(root) == Sequence:
        if seq_field is not None:
            if seq_field not in list(root.annotations.keys()):
                print(
                    "\nERROR: {} is not present in the supplied root sequence.\n".format(
                        seq_field
                    )
                )
                sys.exit(1)
            root.alignment_sequence = root[seq_field]
        if name_field is not None:
            if name_field not in list(root.annotations.keys()):
                print(
                    "\nERROR: {} is not present in the supplied root sequence.\n".format(
                        name_field
                    )
                )
                sys.exit(1)
            root.alignment_id = root[name_field]
        sequences = [s for s in sequences if s.alignment_id != root.alignment_id]
    else:
        print(
            "\nERROR: If root is provided, it must be the name of a sequence \
              found in the supplied list of sequences or it must be a Sequence object."
        )
        print("\n")
        sys.exit(1)

    if root_name is not None:
        root.alignment_id = root_name
    else:
        root_name = root.alignment_id

    # compute and parse the alignment
    seqs = [(root.alignment_id, root.alignment_sequence)]
    seqs += [(s.alignment_id, s.alignment_sequence) for s in sequences]
    aln = muscle(seqs)
    g_aln = [a for a in aln if a.id == root_name][0]
    dots = [
        (root_name, str(g_aln.seq)),
    ]
    for seq in [a for a in aln if a.id != root_name]:
        s_aln = ""
        for g, q in zip(str(g_aln.seq), str(seq.seq)):
            if g == q == "-":
                s_aln += "-"
            elif g == q:
                s_aln += "."
            else:
                s_aln += q
        dots.append((seq.id, s_aln))
    if just_alignment:
        return [d[1] for d in dots]
    name_len = max([len(d[0]) for d in dots]) + 2
    dot_aln = []
    for d in dots:
        if as_fasta:
            dot_aln.append(">{}\n{}".format(d[0], d[1]))
        else:
            spaces = name_len - len(d[0])
            dot_aln.append(d[0] + " " * spaces + d[1])
    return "\n".join(dot_aln)



#!/usr/bin/env python
# filename: base.py


#
# Copyright (c) 2017 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


from __future__ import print_function

import numpy as np
import pandas as pd

import matplotlib.pyplot as plt
import seaborn as sns


__all__ = ['barplot', 'heatmap']


def barplot(x, y, colors, fig_file=None, xlabel=None, ylabel=None,
            rotate_xtick_labels=False, grid=False, size=None, xfontsize=None):
    # set bar locations and width
    ind = np.arange(len(x))
    width = 0.75
    # plot objects
    if size:
        fig = plt.figure(figsize=size)
    else:
        fig = plt.figure()
    ax = fig.add_subplot(111)
    # axis limits and ticks
    ax.set_ylim(0, 1.05 * max(y))
    ax.set_xlim(-width, len(ind) - (width / 2))
    # ax.set_xticks(ind + width / 2)
    ax.set_xticks(ind)
    xtick_names = ax.set_xticklabels(x)
    if rotate_xtick_labels:
        plt.setp(xtick_names, rotation=90, fontsize=7)
    if grid:
        ax.yaxis.grid(True, alpha=0.5)
    # axis labels
    if xlabel is not None:
        ax.set_xlabel(xlabel)
    if ylabel is not None:
        ax.set_ylabel(ylabel)
    if xfontsize is not None:
        plt.setp(xtick_names, fontsize=xfontsize)
    ax.tick_params(axis='x', which='both', top='off', length=3, pad=1.5)
    ax.tick_params(axis='y', which='both', right='off', length=3, pad=1.5)
    # make the plot
    ax.bar(ind, y, width, color=colors)
    fig.tight_layout()
    if fig_file is not None:
        plt.savefig(fig_file)
    else:
        plt.show()
    plt.close()



def heatmap(df, fig_file=None):
    sns.set()
    # set up plot, determine plot size
    h, w = df.shape
    f, ax = plt.subplots(figsize=(w / 1.75, h / 3))
    sns.heatmap(df,
                square=True,
                cmap='Blues',
                cbar=True,
                cbar_kws={'orientation': 'horizontal',
                          'fraction': 0.02,
                          'pad': 0.02,
                          'shrink': 0.675})
    # adjust labels
    ax.xaxis.tick_top()
    plt.xticks(rotation=90)
    f.tight_layout()
    # make the plot
    if fig_file is not None:
        plt.savefig(fig_file)
    else:
        plt.show()
    plt.close()


#!/usr/bin/env python
# filename: scatter.py


#
# Copyright (c) 2022 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


import itertools
from typing import Iterable, Optional, Union

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from natsort import natsorted

from ..core.sequence import Sequence
from ..utils.color import get_cmap, true_false_palette
from .data import process_input_data
from .utils import get_inset_axes_bounds

__all__ = ["scatter"]


def scatter(
    x: Union[str, Iterable],
    y: Union[str, Iterable],
    hue: Union[str, Iterable, None] = None,
    marker: str = "o",
    data: Optional[pd.DataFrame] = None,
    sequences: Optional[Iterable[Sequence]] = None,
    hue_order: Optional[Iterable] = None,
    force_categorical_hue: bool = False,
    force_continuous_hue: bool = False,
    palette: Union[dict, Iterable, None] = None,
    color: Union[str, Iterable, None] = None,
    cmap: Union[str, mpl.colors.Colormap, None] = None,
    hue_min: Optional[float] = None,
    hue_max: Optional[float] = None,
    under_color: Union[str, Iterable, None] = "whitesmoke",
    size: Union[int, float] = 20,
    alpha: float = 1.0,
    highlight_index: Optional[Iterable] = None,
    highlight_x: Optional[Iterable] = None,
    highlight_y: Optional[Iterable] = None,
    highlight_marker: str = "x",
    highlight_size: Union[int, float] = 90,
    highlight_color: Union[str, Iterable] = "k",
    highlight_name: Optional[str] = None,
    highlight_alpha: float = 0.9,
    plot_kwargs: Optional[dict] = None,
    legend_marker_alpha: Optional[float] = None,
    legend_marker_scale: Optional[float] = None,
    legend_on_data: bool = False,
    legend_fontsize: Union[int, float] = 12,
    legend_fontweight: str = "bold",
    legend_fontoutline: Union[int, float] = 3,
    legend_label_position_offsets: Optional[dict] = None,
    legend_kwargs: Optional[dict] = None,
    hide_legend: bool = False,
    xlabel: Optional[str] = None,
    ylabel: Optional[str] = None,
    title: Optional[str] = None,
    title_fontsize: Union[int, float] = 20,
    title_fontweight: str = "normal",
    title_loc: str = "center",
    title_pad: Union[int, float, None] = None,
    show_title: bool = False,
    xlabel_fontsize: Union[int, float] = 16,
    ylabel_fontsize: Union[int, float] = 16,
    xtick_labelsize: Union[int, float] = 14,
    ytick_labelsize: Union[int, float] = 14,
    xtick_labelrotation: Union[int, float] = 0,
    ytick_labelrotation: Union[int, float] = 0,
    hide_ticks: bool = False,
    tiny_axis: bool = False,
    tiny_axis_xoffset: Union[int, float, None] = None,
    tiny_axis_yoffset: Union[int, float, None] = None,
    cbar_width: float = 0.35,
    cbar_height: float = 0.05,
    cbar_loc: str = "lower right",
    cbar_orientation: str = "horizontal",
    cbar_bbox_to_anchor: Optional[Iterable] = None,
    cbar_flip_ticks: bool = False,
    cbar_title: Optional[str] = None,
    cbar_title_fontsize: Union[int, float] = 12,
    cbar_title_loc: Optional[str] = None,
    cbar_title_labelpad: float = 8.0,
    hide_cbar: bool = False,
    equal_axes: bool = True,
    ax: Optional[mpl.axes.Axes] = None,
    show: bool = False,
    figsize: Optional[Iterable] = None,
    figfile: Optional[str] = None,
) -> Optional[mpl.axes.Axes]:
    """
    Produces a scatter plot.

    Parameters
    ----------
    x : str or iterable object
        Name of a column in `data` or an iterable of values to be plotted on the
        x-axis. Required.

    y : str or iterable object
        Name of a column in `data` or an iterable of values to be plotted on the
        y-axis. Required.

    hue : str or iterable object, optional
        Name of a column in `data` or an iterable of hue categories.

    marker : str, dict or iterable object, optional
        Marker style for the scatter plot. Accepts any of the following:
          * a `matplotlib marker`_ string
          * a ``dict`` mapping `hue` categories to a `matplotlib marker`_ string
          * a ``list`` of `matplotlib marker`_ strings, which should be the same
              length as `x` and `y`.

    data : pandas.DataFrame, optional
        A ``DataFrame`` object containing the input data. If provided, `x`, `y` and/or `hue` should
        be column names in `data`.

    sequences : iterable of abutils.core.sequence.Sequence, optional
        An iterable of ``Sequence`` objects. If provided, `x`, `y` and/or `hue` should be annotations
        in the ``Sequence`` objects. Alternatively, `x`, `y` and/or `hue` can be an iterable of
        values to be plotted, but must be the same length as `sequences`.

    hue_order : iterable object, optional
        List of `hue` categories in the order they should be plotted. If `hue_order` contains a
        subset of all categories found in `hue`, only the supplied categories will be plotted.
        If not provided, `hue` categories will be plotted in ``natsort.natsorted()`` order.

    force_categorical_hue : bool, default=False
        If ``True``, `hue` data will be treated as categorical, even if the data appear to
        be continuous. This results in `color` being used to color the points rather than `cmap`.

    force_continuous_hue : bool, default=False
        If ``True``, `hue` data will be treated as continuous, even if the data appear to
        be categorical. This results in `cmap` being used to color the points rather than `color`.
        This may produce unexpected results and/or errors if used on non-numerical data.

    palette : dict, optional
        Dictionary mapping `hue`, `x` or `y` names to colors. If if keys in `palette` match
        more than one category, `hue` categories take priority. If `palette` is not provided,
        bars are colored using `color` (if `hue` is ``None``) or a palette is generated
        automatically using ``sns.hls_palette()``.

    color : str or iterable object, optional.
        Color for the plot markers. Can be any of the following:
          * a ``list`` or ``tuple`` containing RGB or RGBA values
          * a color string, either from `Matplotlib's set of named colors`_ or a hex color value
          * the name of a column in `data` that contains color values
          * a ``list`` of colors (either as strings or RGB tuples) to be used for `hue` categories.
            If `colors` is shorter than the list of hue categories, colors will be reused.

        .. tip::
            If a single RGB or RGBA ``list`` or ``tuple`` is provided and `hue` is also supplied,
            there may be unexpected results as ``scatter()`` will attempt to map each of the
            individual RGB(A) values to a hue category. Wrapping the RGB(A) iterable in a list
            will ensure that the color is interpreted correctly.

        Only used if `hue` contains categorical data (`cmap` is used for continuous data). If not
        provided, the `default Seaborn color palette`_ will be used.

    cmap : str or matplotlib.color.Colormap, default='flare'
        Colormap to be used for continuous `hue` data.

    hue_min : float, default=0
        Minimum value for `hue` when `hue` is continuous. Values below `hue_min` will be set to
        `under_color`.

    hue_max : float, default=1
        Maximum value for `hue` when `hue` is continuous. Values at or above `hue_max` will
        all be colored as the maxmum value in `cmap`.

    under_color : str or list of RGB(A) values
        Separate color for ``0`` values when `hue` is continuous. By default, `cmap` is
        used for all values. An example use would be GEx plots for which visualization is
        improved if ``0`` values are more obviously distinguished from low count values.

    size : str or float or iterable object, default=20
        Size of the scatter points. Either a

    alpha : float, default=0.6
        Alpha of the scatter points.

    highlight_index : iterable object, optional
        An iterable of index names (present in `data.index`) of points to be highlighted on
        the plot. If provided, `highlight_x` and `highlight_y` are ignored.

    highlight_x : iterable object, optional
        An iterable of x-values for highlighted points. Also requires `highlight_y`.

    highlight_y : iterable object, optional
        An iterable of y-values for highlighted points. Also requires `highlight_x`.

    highlight_marker : str, default='x'
        Marker style to be used for highlight points. Accepts any `matplotlib marker`_.

    highlight_size : int, default=90
        Size of the highlight marker.

    highlight_color : string or list of color values, default='k'
        Color of the highlight points.

    highlight_name : str, optional
        Name of the highlights, to be used in the plot legend. If not supplied,
        highlight points will not be included in the legend.

    highlight_alpha : float, default=0.9
        Alpha of the highlight points.

    plot_kwargs : dict, optional
        Dictionary containing keyword arguments that will be passed to ``pyplot.scatter()``.

    legend_on_data : bool, default=False
        Plot legend labels on the data rather than in a separate legend. The X/Y midpoint
        for each legend category is used as the label location.

    legend_fontsize : int or float, default=12
        Fontsize for legend labels.

    legend_fontweight : str, default="normal"
        Fontweight for legend labels. Options are any accepted `Matplotlib text weight`_.

    legend_fontoutline : int or float, default=None
        Width of the outline of legend labels. Only used when `legend_on_data` is ``True``.
        Default is ``None``, which results in no outline.

    legend_marker_alpha : float, default=None
        Opacity for legend markers (or legend labels if `legend_on_data` is ``True``).
        By default, legend markers will use `alpha` and legend labels will be completely
        opaque, equivalent to `legend_marker_alpha` of ``1.0``.

    legend_marker_scale : float, default=None
        Scale for legend markers.

    legend_label_position_offsets : dict, default=None
        A ``dict`` mapping legend labels to ``(x,y)`` coordinates used to offset legend labels.
        Only used when `legend_on_data` is ``True``. Offsets are in relative plot units: ``(0.1, 0.1)``
        would move the label up and to the right by 10% of the overall plot area.

    legend_kwargs : dict, optional
        Dictionary containing keyword arguments that will be passed to ``ax.legend()``.

    hide_legend : bool, default=False
        By default, a plot legend will be shown if multiple hues are plotted. If ``True``,
        the legend will not be shown.

    xlabel : str, optional
        Text for the X-axis label.

    ylabel : str, optional
        Text for the Y-axis label.

    xlabel_fontsize : int or float, default=16
        Fontsize for the X-axis label text.

    ylabel_fontsize : int or float, default=16
        Fontsize for the Y-axis label text.

    xtick_labelsize : int or float, default=14
        Fontsize for the X-axis tick labels.

    ytick_labelsize : int or float, default=14
        Fontsize for the Y-axis tick labels.

    xtick_labelrotation : int or float, default=0
        Rotation of the X-axis tick labels.

    ytick_labelrotation : int or float, default=0
        Rotation of the Y-axis tick labels.

    tiny_axis : bool, default=False
        Plots tiny axis lines in the lower left corner of the plot. Typcally used in
        UMAP plots. If ``True``, ticks and tick labels will be hidden.

    tiny_axis_xoffset : float, default=None
        X-axis offset for the tiny axis.

    tiny_axis_yoffset : float, default=None
        Y-axis offset for the tiny axis.

    cbar_width : int, default=35
        Width of the colorbar. Only used for continuous `hue` types.

    cbar_height : int, default=5
        Height of the colorbar. Only used for continuous `hue` types.

    cbar_loc : str or iterable object, default='lower right'
        Location of the colorbar. Accepts `any valid inset_axes() location`_.

    cbar_orientation : str, default='horizontal'
        Orientation of the colorbar. Options are ``'horizontal'`` and ``'vertical'``.

    cbar_bbox_to_anchor : list or tuple, optional
        bbox_to_anchor for the colorbar. Used in combination with `cbar_loc` to provide
        fine-grained positioning of the colorbar.

    cbar_flip_ticks : bool, default=False
        Flips the position of colorbar ticks. Ticks default to the bottom if `cbar_orientation`
        is  ``'horizontal'`` and the left if  `cbar_orientation` is ``'vertical'``.

    cbar_title : str, optional
        Colorbar title. If not provided, `hue` is used.

    cbar_title_fontsize : int or float, default=12
        Fontsize for the colorbar title.

    hide_cbar : bool, default=False.
        If ``True``, the color bar will be hidden on plots with continuous `hue` values.

    equal_axes : bool, default=True
        If ```True```, the the limits of the x- and y-axis will be equal.

    ax : mpl.axes.Axes, default=None
        Pre-existing axes for the plot. If not provided, a new axes will be created.

    show : bool, default=False
        If ``True``, plot is shown and the plot ``Axes`` object is not returned. Default
        is ``False``, which does not call ``pyplot.show()`` and returns the ``Axes`` object.

    figsize : iterable object, default=[6, 4]
        List containing the figure size (as ``[x-dimension, y-dimension]``) in inches.

    figfile : str, optional
        Path at which to save the figure file. If not provided, the figure is not saved
        and is either shown (if `show` is ``True``) or the ``Axes`` object is returned.


    Returns
    -------
    ax : mpl.axes.Axes
        If `figfile` is ``None`` and `show` is ``False``, the ``ax`` is returned.
        Otherwise, the return value is ``None``.


    .. _matplotlib marker:
        https://matplotlib.org/stable/api/markers_api.html

    .. _Matplotlib's set of named colors:
        https://matplotlib.org/stable/gallery/color/named_colors.html

    .. _default Seaborn color palette:
        https://seaborn.pydata.org/generated/seaborn.color_palette.html

    .. _Matplotlib text weight
        https://matplotlib.org/stable/tutorials/text/text_props.html

    .. _any valid inset_axes() location:
        https://matplotlib.org/stable/api/_as_gen/mpl_toolkits.axes_grid1.inset_locator.inset_axes.html

    """
    # process input data
    # if data is None:
    #     _data = {}
    #     _data["x"] = x
    #     x = "x"
    #     _data["y"] = y
    #     y = "y"
    #     if hue is not None:
    #         _data["hue"] = hue
    #         hue = "hue"
    #     df = pd.DataFrame(_data)
    # else:
    #     df = data.copy()
    #     if not isinstance(x, str) and len(x) == df.shape[0]:
    #         df["x"] = x
    #         x = "x"
    #     if not isinstance(y, str) and len(y) == df.shape[0]:
    #         df["y"] = y
    #         y = "y"
    #     if hue is not None:
    #         if not isinstance(hue, str) and len(hue) == df.shape[0]:
    #             df["hue"] = hue
    #             hue = "hue"
    df, x, y, hue = process_input_data(
        x=x, y=y, hue=hue, data=data, sequences=sequences
    )

    # figure size
    if figsize is None:
        figsize = [6, 6]

    # hue and color
    continuous_hue = False
    if hue is not None:
        if force_continuous_hue:
            continuous_hue = True
        elif all([isinstance(h, float) for h in df[hue]]) and not force_categorical_hue:
            continuous_hue = True
        if continuous_hue:
            continuous_hue = True
            # set hue min and max values
            if hue_min is None:
                hue_min = np.floor(df[hue].min())
            if hue_max is None:
                hue_max = np.ceil(df[hue].max())
            hue_order = []
            cmap = get_cmap("flare" if cmap is None else cmap)
            normhue = lambda h: (h - hue_min) / (hue_max - hue_min)
            df["color"] = [
                cmap(normhue(h)) if h >= hue_min else under_color for h in df[hue]
            ]
        else:
            if hue_order is None:
                hue_order = natsorted(list(set(df[hue])))
            if palette is not None:
                missing_color = color if color is not None else "lightgrey"
                hue_dict = {h: palette.get(h, missing_color) for h in hue_order}
            else:
                if all([isinstance(h, bool) for h in hue_order]):
                    hue_dict = true_false_palette
                else:
                    n_colors = max(1, len(hue_order))
                    if color is None:
                        color = sns.color_palette(n_colors=n_colors)
                    if len(color) < n_colors:
                        color = itertools.cycle(color)
                    hue_dict = {h: c for h, c in zip(hue_order, color)}
            df["color"] = [hue_dict[h] for h in df[hue]]
    else:
        hue_order = []
        if isinstance(color, str) and color in df.columns:
            df["color"] = df[color]
        elif color is not None:
            df["color"] = [color] * df.shape[0]
        else:
            df["color"] = [sns.color_palette()[0]] * df.shape[0]

    # markers
    # TODO: allow different markers that correspond to marker "categories"
    # much like we can do with hues
    # this might be a bit complex, because pyplot does not currently
    # support marker style assignment by list, so we'd need to make
    # a separate plt.scatter() call for each marker style

    # plot kwargs
    default_plot_kwargs = {"linewidths": 0}
    if plot_kwargs is not None:
        default_plot_kwargs.update(plot_kwargs)
    plot_kwargs = default_plot_kwargs

    # scatterplot
    if ax is None:
        plt.figure(figsize=figsize)
        ax = plt.gca()
    if hue_order:
        for h in hue_order[::-1]:
            d = df[df[hue] == h]
            ax.scatter(
                d[x],
                d[y],
                c=d["color"],
                s=size,
                marker=marker,
                alpha=alpha,
                label=h,
                **plot_kwargs,
            )
    else:
        ax.scatter(
            df[x],
            df[y],
            c=df["color"],
            s=size,
            marker=marker,
            alpha=alpha,
            **plot_kwargs,
        )

    # highlighted points
    highlight = any(
        [
            highlight_index is not None,
            all([highlight_x is not None, highlight_y is not None]),
        ]
    )
    if highlight:
        if highlight_index is not None:
            hi_index = [h for h in highlight_index if h in df.index.values]
            hi_df = df.loc[hi_index]
            highlight_x = hi_df[x]
            highlight_y = hi_df[y]
        plt.scatter(
            highlight_x,
            highlight_y,
            zorder=10,
            s=highlight_size,
            c=highlight_color,
            alpha=highlight_alpha,
            marker=highlight_marker,
            label=highlight_name,
        )

    # legend
    if not continuous_hue:
        if hue is not None and not hide_legend:
            default_legend_kwargs = {"frameon": True, "loc": "best", "fontsize": 12}
            if legend_kwargs is not None:
                default_legend_kwargs.update(legend_kwargs)
            if legend_marker_scale is not None:
                default_legend_kwargs["markerscale"] = legend_marker_scale
            legend_kwargs = default_legend_kwargs
            ax.legend(**legend_kwargs)
            if legend_marker_alpha is not None:
                leg = ax.get_legend()
                for lh in leg.legendHandles:
                    lh.set_alpha(legend_marker_alpha)

    # colorbar
    if continuous_hue and not hide_cbar:
        if cbar_orientation == "horizontal":
            width = max([cbar_width, cbar_height])
            height = min([cbar_width, cbar_height])
        else:
            width = min([cbar_width, cbar_height])
            height = max([cbar_width, cbar_height])
        cbar_bounds = get_inset_axes_bounds(
            cbar_loc, cbar_bbox_to_anchor, width, height
        )
        cbax = ax.inset_axes(cbar_bounds)

        # max_hue = np.ceil(df[hue].max())
        # min_hue = max(0, df[hue].min())
        norm = mpl.colors.Normalize(vmin=hue_min, vmax=hue_max)
        # ticks = [t for t in np.linspace(min_hue, max_hue, num=4)]

        cbar = plt.colorbar(
            mpl.cm.ScalarMappable(norm=norm, cmap=cmap),
            cax=cbax,
            orientation=cbar_orientation,
        )
        if cbar_orientation == "horizontal":
            if cbar_title_loc is None:
                cbar_title_loc = "top"
            ticks_position = "bottom" if cbar_flip_ticks else "top"
            cbax.xaxis.set_ticks_position(ticks_position)
            cbax.xaxis.set_label_position(cbar_title_loc)
            cbar.ax.set_xlabel(
                cbar_title,
                fontsize=cbar_title_fontsize,
                labelpad=cbar_title_labelpad,
            )
        else:
            if cbar_title_loc is None:
                cbar_title_loc = "right"
            ticks_position = "left" if cbar_flip_ticks else "right"
            cbax.yaxis.set_ticks_position(ticks_position)
            cbax.yaxis.set_label_position(cbar_title_loc)
            cbar.ax.set_ylabel(
                cbar_title,
                fontsize=cbar_title_fontsize,
                labelpad=cbar_title_labelpad,
            )

    # style the plot
    ax.set_xlabel(xlabel if xlabel is not None else x, fontsize=xlabel_fontsize)
    ax.set_ylabel(ylabel if ylabel is not None else y, fontsize=ylabel_fontsize)
    ax.tick_params(
        axis="x", labelsize=xtick_labelsize, labelrotation=xtick_labelrotation
    )
    ax.tick_params(
        axis="y", labelsize=ytick_labelsize, labelrotation=ytick_labelrotation
    )

    if tiny_axis:
        # get coords for the UMAP-specific axes
        if tiny_axis_xoffset is None:
            tiny_axis_xoffset = -0.05
        if tiny_axis_yoffset is None:
            tiny_axis_yoffset = -0.05
        xmin = df[x].min()
        xmax = df[x].max()
        ymin = df[y].min()
        ymax = df[y].max()
        x_range = abs(xmax - xmin)
        y_range = abs(ymax - ymin)
        x_offset = x_range * tiny_axis_xoffset
        y_offset = y_range * tiny_axis_yoffset
        x_start = xmin + x_offset
        y_start = ymin + y_offset
        x_end = xmin + (x_range / 5) + x_offset
        x_center = x_start + ((x_end - x_start) / 2)
        y_end = ymin + (y_range / 5) + y_offset
        y_center = y_start + ((y_end - y_start) / 2)
        # draw the new "mini" axis lines
        ax.hlines(y_start, x_start, x_end, "k", lw=2)
        ax.vlines(x_start, y_start, y_end, "k", lw=2)
        ax.annotate(
            xlabel,
            xy=(x_center, y_start),
            xytext=(0, -5),
            textcoords="offset points",
            fontsize=xlabel_fontsize,
            ha="center",
            va="top",
        )
        ax.annotate(
            ylabel,
            xy=(x_start, y_center),
            xytext=(-5, 0),
            textcoords="offset points",
            fontsize=ylabel_fontsize,
            rotation="vertical",
            ha="right",
            va="center",
        )
        # remove the normal axis lines
        ax.set_xlabel("", fontsize=0)
        ax.set_ylabel("", fontsize=0)
        for s in ["left", "right", "top", "bottom"]:
            ax.spines[s].set_visible(False)
    else:
        for spine in ["right", "top"]:
            ax.spines[spine].set_visible(False)
        for spine in ["left", "bottom"]:
            ax.spines[spine].set_position(("outward", 10))

    if all([hue is not None, legend_on_data, not continuous_hue]):
        # set up label offsets
        if legend_label_position_offsets is None:
            legend_label_position_offsets = {}
        _xlim = ax.get_xlim()
        _ylim = ax.get_ylim()
        _xrange = _xlim[1] - _xlim[0]
        _yrange = _ylim[1] - _ylim[0]
        # configure the legend font outline
        if legend_fontoutline is not None:
            path_effect = [
                mpl.patheffects.withStroke(linewidth=legend_fontoutline, foreground="w")
            ]
        else:
            path_effect = None
        # add the on-data legend
        for h in df[hue].unique():
            _df = df[df[hue] == h]
            xoffset, yoffset = legend_label_position_offsets.get(h, (0, 0))
            hue_x = _df[x].median() + (xoffset * _xrange)
            hue_y = _df[y].median() + (yoffset * _yrange)
            ax.text(
                hue_x,
                hue_y,
                h,
                c="k",
                alpha=legend_marker_alpha,
                weight=legend_fontweight,
                verticalalignment="center",
                horizontalalignment="center",
                fontsize=legend_fontsize,
                path_effects=path_effect,
            )
        # hide the "real" legend
        hide_legend = True

    if hide_legend:
        l = ax.get_legend()
        if l is not None:
            l.remove()

    if hide_ticks or tiny_axis:
        ax.set_xticks([])
        ax.set_yticks([])

    if equal_axes:
        xlim = ax.get_xlim()
        ylim = ax.get_ylim()
        axlim = [min([xlim[0], ylim[0]]), max([xlim[1], ylim[1]])]
        ax.set_xlim(axlim)
        ax.set_ylim(axlim)

    if show_title and title is not None:
        ax.set_title(
            title,
            loc=title_loc,
            pad=title_pad,
            fontsize=title_fontsize,
            fontweight=title_fontweight,
        )

    # save, show or return the ax
    if figfile is not None:
        plt.tight_layout()
        plt.savefig(figfile)
    elif show:
        plt.show()
    else:
        return ax


# def get_inset_axes_bounds(loc, bbox_to_anchor, width, height):
#     if bbox_to_anchor is None:
#         loc_dict = {
#             "upper left": [0, 1 - height],
#             "upper center": [0.5 - width / 2, 1 - height],
#             "upper right": [1 - width, 1 - height],
#             "center left": [0, 0.5 - height / 2],
#             "center": [0.5 - width / 2, 0.5 - height / 2],
#             "center right": [1 - width, 0.5 - height / 2],
#             "lower left": [0, 0],
#             "lower center": [0.5 - width / 2, 0],
#             "lower right": [1 - width, 0],
#         }
#         x0, y0 = loc_dict.get(loc, [0, 0])
#     else:
#         x, y = bbox_to_anchor[:2]
#         loc_dict = {
#             "upper left": [x, y - height],
#             "upper center": [x - width / 2, y - height],
#             "upper right": [x - width, y - height],
#             "center left": [x, y - height / 2],
#             "center": [x - width / 2, y - height / 2],
#             "center right": [x - width, y - height / 2],
#             "lower left": [x, y],
#             "lower center": [x - width / 2, y],
#             "lower right": [x - width, y],
#         }
#         x0, y0 = loc_dict.get(loc, [0, 0])
#     return [x0, y0, width, height]



# import os
# import sys

# try:
#     from setuptools import setup
# except ImportError:
#     sys.exit('ERROR: setuptools is required.\n')


# try: # for pip >= 10
#     from pip._internal.req import parse_requirements
# except ImportError: # for pip <= 9.0.3
#     from pip.req import parse_requirements
# # try:
# #     from pip.req import parse_requirements
# # except ImportError:
# #     sys.exit('ERROR: pip is required.\n')


# if os.environ.get('READTHEDOCS', None):
#     # Set empty install_requires to get install to work on readthedocs
#     install_requires = []
# else:
#     if sys.version_info[0] > 2:
#         req_file = 'requirements.txt'
#     else:
#         req_file = 'requirements2.txt'
#     try:
#         reqs = parse_requirements(req_file, session=False)
#     except TypeError:
#         reqs = parse_requirements(req_file)
#     try:
#         install_requires = [str(r.req) for r in reqs]
#     except AttributeError:
#         install_requires = [str(r.requirement) for r in reqs]

# # read version
# exec(open('abutils/version.py').read())

# config = {
#     'description': 'Utilities for analysis of antibody NGS data',
#     'author': 'Bryan Briney',
#     'url': 'https://www.github.com/briney/abutils',
#     'author_email': 'briney@scripps.edu',
#     'version': __version__,
#     'install_requires': install_requires,
#     'packages': ['abutils'],
#     'scripts': [],
#     'name': 'abutils',
#     'include_package_data': True,
#     'classifiers': ['License :: OSI Approved :: MIT License',
#                     'Programming Language :: Python :: 3.6',
#                     'Programming Language :: Python :: 3.7',
#                     'Programming Language :: Python :: 3.8',
#                     'Topic :: Scientific/Engineering :: Bio-Informatics']
# }

# setup(**config)


import os

from setuptools import find_packages, setup

# read version
version_file = os.path.join(os.path.dirname(__file__), "abutils", "version.py")
with open(version_file) as f:
    exec(f.read())

# read requirements
with open("requirements.txt") as f:
    requirements = f.read().splitlines()

# read long description
with open("README.md", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="abutils",
    version=__version__,
    author="Bryan Briney",
    author_email="briney@scripps.edu",
    description="Utilities for analysis of adaptive immune receptor repertoire (AIRR) data",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/briney/abutils",
    packages=find_packages(),
    # package_dir={"": "abutils"},
    include_package_data=True,
    install_requires=requirements,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        # 'Programming Language :: Python :: 3.6',
        # 'Programming Language :: Python :: 3.7',
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    python_requires=">=3.8",
)



#!usr/env/python
# filename: test_sequences.py

import os
import tempfile

import pandas as pd
import polars as pl
import pytest
from Bio.SeqRecord import SeqRecord

from ..core.sequence import Sequence
from ..io import (
    read_airr,
    read_csv,
    read_fasta,
    read_parquet,
    to_fasta,
)


@pytest.fixture
def fasta_sequence():
    f = ">fasta_sequence\nMEEPQSDPSVEPPLSQETFSDLWKLLPENNVLSPLPSQAMDDLMLSPDDIEQWFTEDPGP"
    fasta = tempfile.NamedTemporaryFile(delete=False)
    fasta.write(f.encode("utf-8"))
    fasta.close()
    return fasta


@pytest.fixture
def airr_sequences():
    a = "sequence_id\tsequence\tsequence_aa\nairr_sequence\tATCGATCGATCG\tCASSLDRYFDYW\nairr_sequence2\tATCGATCGATCG\tCASSLDRYFDYW\n"
    airr = tempfile.NamedTemporaryFile(delete=False)
    airr.write(a.encode("utf-8"))
    airr.close()
    return airr.name


@pytest.fixture
def parquet_sequences():
    a = [
        {
            "sequence_id": "airr_sequence",
            "sequence": "ATCGATCGATCG",
            "sequence_aa": "CASSLDRYFDYW",
        },
        {
            "sequence_id": "airr_sequence2",
            "sequence": "ATCGATCGATCG",
            "sequence_aa": "CASSLDRYFDYW",
        },
    ]
    df = pl.DataFrame(a)
    parquet = tempfile.NamedTemporaryFile(delete=False)
    df.write_parquet(parquet.name)
    parquet.close()
    return parquet.name


@pytest.fixture
def csv_sequences():
    a = "sequence_id,sequence,sequence_aa\nairr_sequence,ATCGATCGATCG,CASSLDRYFDYW\nairr_sequence2,ATCGATCGATCG,CASSLDRYFDYW\n"
    csv = tempfile.NamedTemporaryFile(delete=False)
    csv.write(a.encode("utf-8"))
    csv.close()
    return csv.name


@pytest.fixture
def polars_sequences():
    a = [
        {
            "sequence_id": "airr_sequence",
            "sequence": "ATCGATCGATCG",
            "sequence_aa": "CASSLDRYFDYW",
        },
        {
            "sequence_id": "airr_sequence2",
            "sequence": "ATCGATCGATCG",
            "sequence_aa": "CASSLDRYFDYW",
        },
    ]
    df = pl.DataFrame(a)
    return df


@pytest.fixture
def pandas_sequences():
    a = [
        {
            "sequence_id": "airr_sequence",
            "sequence": "ATCGATCGATCG",
            "sequence_aa": "CASSLDRYFDYW",
        },
        {
            "sequence_id": "airr_sequence2",
            "sequence": "ATCGATCGATCG",
            "sequence_aa": "CASSLDRYFDYW",
        },
    ]
    df = pd.DataFrame(a)
    return df


@pytest.fixture
def sequences():
    seq1 = Sequence(
        {
            "sequence": "ATCG",
            "seq": "ATCG",
            "sequence_id": "seq1",
            "name": "seq1",
        }
    )
    seq2 = Sequence(
        {
            "sequence": "GCTA",
            "seq": "GCTA",
            "sequence_id": "seq2",
            "name": "seq2",
        }
    )
    seq3 = Sequence(
        {
            "sequence": "TTTT",
            "seq": "TTTT",
            "sequence_id": "seq3",
            "name": "seq3",
        }
    )
    return [seq1, seq2, seq3]


@pytest.fixture
def seq_records():
    seq_records = [
        SeqRecord("ATCG", id="seq1"),
        SeqRecord("GCTA", id="seq2"),
        SeqRecord("TTTT", id="seq3"),
    ]
    return seq_records


# ------------------
#     Sequence
# ------------------


def test_sequence_with_string():
    seq = Sequence("ATCG")
    assert seq.sequence == "ATCG"
    assert seq.id is not None


def test_sequence_with_string_and_id():
    seq = Sequence("ATCG", id="seq1")
    assert seq.sequence == "ATCG"
    assert seq.id == "seq1"


def test_sequence_with_list():
    seq = Sequence(["seq1", "ATCG"])
    assert seq.sequence == "ATCG"
    assert seq.id == "seq1"


def test_sequence_with_dict():
    seq = Sequence({"sequence_id": "seq1", "sequence": "ATCG"})
    assert seq.sequence == "ATCG"
    assert seq.id == "seq1"


def test_sequence_with_dict_and_keys():
    seq = Sequence({"name": "seq1", "dna": "ATCG"}, id_key="name", seq_key="dna")
    assert seq.sequence == "ATCG"
    assert seq.id == "seq1"


def test_sequence_with_seqrecord():
    from Bio.SeqRecord import SeqRecord

    seq = SeqRecord("ATCG")
    seq.id = "seq1"
    ab_seq = Sequence(seq)
    assert ab_seq.sequence == "ATCG"
    assert ab_seq.id == "seq1"


def test_sequence_with_abutils_sequence():
    seq = Sequence("ATCG")
    ab_seq = Sequence(seq)
    assert ab_seq.sequence == "ATCG"
    assert ab_seq.id is not None


def test_sequence_contains():
    seq = Sequence("ATCG")
    assert "AT" in seq
    assert "CG" in seq
    assert "GC" not in seq


def test_sequence_getitem():
    seq = Sequence("ATCG")
    assert seq[0] == "A"
    assert seq[1:3] == "TC"
    assert seq[-1] == "G"


def test_sequence_fastq():
    seq = Sequence("ATCG", qual="abcd")
    assert seq.fastq == "@{}\n{}\n+\n{}".format(seq.id, seq.sequence, "abcd")


def test_sequence_reverse_complement():
    seq = Sequence("ATCG")
    assert seq.reverse_complement == "CGAT"


def test_sequence_annotations():
    seq = Sequence("ATCG")
    assert seq.annotations == {}
    seq.annotations = {"key": "value"}
    assert seq.annotations == {"key": "value"}


def test_sequence_strand():
    seq = Sequence("ATCG")
    assert seq.strand == "plus"
    seq.strand = "minus"
    assert seq.strand == "minus"


def test_sequence_translate():
    seq = Sequence("ATGTGCTAA")
    assert seq.translate() == "MC*"


def test_sequence_as_fasta():
    seq = Sequence("ATCG", id="seq1", annotations={"name": "my_seq", "seq": "TACG"})
    assert seq.as_fasta() == ">seq1\nATCG"
    assert seq.as_fasta(name_field="name") == ">my_seq\nATCG"
    assert seq.as_fasta(seq_field="seq") == ">seq1\nTACG"


def test_sequence_region():
    seq = Sequence("ATCG")
    assert seq.region() == ">{}\n{}".format(seq.id, seq.sequence)
    assert seq.region(start=1) == ">{}\n{}".format(seq.id, seq.sequence[1:])
    assert seq.region(end=2) == ">{}\n{}".format(seq.id, seq.sequence[:2])
    assert seq.region(start=1, end=2) == ">{}\n{}".format(seq.id, seq.sequence[1:2])


# ------------------
#     readers
# ------------------


def test_read_fasta(fasta_sequence):
    seqs = read_fasta(fasta_sequence.name)
    assert len(seqs) == 1
    assert seqs[0].id == "fasta_sequence"
    assert (
        seqs[0].sequence
        == "MEEPQSDPSVEPPLSQETFSDLWKLLPENNVLSPLPSQAMDDLMLSPDDIEQWFTEDPGP"
    )


def test_read_airr(airr_sequences):
    seqs = read_airr(airr_sequences)
    assert len(seqs) == 2
    assert seqs[0].id == "airr_sequence"
    assert seqs[0].sequence == "ATCGATCGATCG"
    assert seqs[0]["sequence_aa"] == "CASSLDRYFDYW"
    assert seqs[1].id == "airr_sequence2"
    assert seqs[1].sequence == "ATCGATCGATCG"
    assert seqs[1]["sequence_aa"] == "CASSLDRYFDYW"


def test_read_parquet(parquet_sequences):
    seqs = read_parquet(parquet_sequences)
    assert len(seqs) == 2
    assert seqs[0].id == "airr_sequence"
    assert seqs[0].sequence == "ATCGATCGATCG"
    assert seqs[0]["sequence_aa"] == "CASSLDRYFDYW"
    assert seqs[1].id == "airr_sequence2"
    assert seqs[1].sequence == "ATCGATCGATCG"
    assert seqs[1]["sequence_aa"] == "CASSLDRYFDYW"


def test_read_csv(csv_sequences):
    seqs = read_csv(csv_sequences)
    assert len(seqs) == 2
    assert seqs[0].id == "airr_sequence"
    assert seqs[0].sequence == "ATCGATCGATCG"
    assert seqs[0]["sequence_aa"] == "CASSLDRYFDYW"
    assert seqs[1].id == "airr_sequence2"
    assert seqs[1].sequence == "ATCGATCGATCG"
    assert seqs[1]["sequence_aa"] == "CASSLDRYFDYW"


# def test_from_polars(polars_sequences):
#     seqs = from_polars(polars_sequences)
#     assert len(seqs) == 2
#     assert seqs[0].id == "airr_sequence"
#     assert seqs[0].sequence == "ATCGATCGATCG"
#     assert seqs[0]["sequence_aa"] == "CASSLDRYFDYW"
#     assert seqs[1].id == "airr_sequence2"
#     assert seqs[1].sequence == "ATCGATCGATCG"
#     assert seqs[1]["sequence_aa"] == "CASSLDRYFDYW"


# def test_from_pandas(pandas_sequences):
#     seqs = from_pandas(pandas_sequences)
#     assert len(seqs) == 2
#     assert seqs[0].id == "airr_sequence"
#     assert seqs[0].sequence == "ATCGATCGATCG"
#     assert seqs[0]["sequence_aa"] == "CASSLDRYFDYW"
#     assert seqs[1].id == "airr_sequence2"
#     assert seqs[1].sequence == "ATCGATCGATCG"
#     assert seqs[1]["sequence_aa"] == "CASSLDRYFDYW"


# ------------------
#    to fasta
# ------------------


def test_to_fasta_with_sequence_objects(sequences):
    fasta = to_fasta(sequences, as_string=True)
    assert isinstance(fasta, str)
    assert ">seq1\nATCG\n>seq2\nGCTA\n>seq3\nTTTT" in fasta


def test_to_fasta_with_fasta_string():
    fasta_string = ">seq1\nATCG\n>seq2\nGCTA\n>seq3\nTTTT"
    fasta = to_fasta(fasta_string, as_string=True)
    assert isinstance(fasta, str)
    assert fasta == fasta_string


def test_to_fasta_with_fasta_file(sequences):
    fasta_string = "\n".join([s.fasta for s in sequences])
    fasta_file = tempfile.NamedTemporaryFile(delete=False)
    fasta_file.write(fasta_string.encode("utf-8"))
    fasta_file.close()
    fasta = to_fasta(fasta_file.name, as_string=True)
    assert isinstance(fasta, str)
    assert ">seq1\nATCG\n>seq2\nGCTA\n>seq3\nTTTT" in fasta


def test_to_fasta_with_seqrecord_objects(seq_records):
    fasta = to_fasta(seq_records, as_string=True)
    assert isinstance(fasta, str)
    assert ">seq1\nATCG\n>seq2\nGCTA\n>seq3\nTTTT" in fasta


def test_to_fasta_with_list_of_lists():
    seqs = [
        ["seq1", "ATCG"],
        ["seq2", "GCTA"],
        ["seq3", "TTTT"],
    ]
    fasta = to_fasta(seqs, as_string=True)
    assert isinstance(fasta, str)
    assert ">seq1\nATCG\n>seq2\nGCTA\n>seq3\nTTTT" in fasta


def test_to_fasta_with_id_key(sequences):
    fasta = to_fasta(sequences, id_key="name", as_string=True)
    assert isinstance(fasta, str)
    assert ">seq1\nATCG\n>seq2\nGCTA\n>seq3\nTTTT" in fasta


def test_to_fasta_with_sequence_key(sequences):
    fasta = to_fasta(sequences, sequence_key="seq", as_string=True)
    assert isinstance(fasta, str)
    assert ">seq1\nATCG\n>seq2\nGCTA\n>seq3\nTTTT" in fasta


def test_to_fasta_with_fasta_file_output(sequences):
    with tempfile.NamedTemporaryFile(delete=False) as fasta_file:
        fasta = to_fasta(sequences, fasta_file=fasta_file.name)
        assert isinstance(fasta, str)
        assert fasta == fasta_file.name
        with open(fasta_file.name, "r") as f:
            fasta_string = f.read()
        assert ">seq1\nATCG\n>seq2\nGCTA\n>seq3\nTTTT" in fasta_string


def test_to_fasta_with_fasta_string_output(sequences):
    fasta = to_fasta(sequences, as_string=True)
    assert isinstance(fasta, str)
    assert ">seq1\nATCG\n>seq2\nGCTA\n>seq3\nTTTT" in fasta


# -------------------------------------------
#          codon optimization
# -------------------------------------------


def test_codon_optimize_returns_string():
    seq = Sequence("RINEYLA")
    optimized_seq = seq.codon_optimize(as_string=True)
    assert isinstance(optimized_seq, str)
    assert len(optimized_seq) == len(seq) * 3


def test_codon_optimize_returns_sequence_object():
    seq = Sequence("RINEYLA")
    optimized_seq = seq.codon_optimize(as_string=False)
    assert isinstance(optimized_seq, Sequence)
    assert len(optimized_seq.sequence) == len(seq) * 3
    assert optimized_seq.translate() == seq.sequence


def test_codon_optimize_with_frame():
    seq = Sequence("ATGCATGCATGCATGC")
    with pytest.raises(ValueError):
        seq.codon_optimize(frame=4)
